"""Concatenate operation module."""

import dataclasses
import os

import pandas as pd

from mlcroissant._src.core.path import Path
from mlcroissant._src.operation_graph.base_operation import Operation
from mlcroissant._src.structure_graph.nodes.file_set import FileSet
from mlcroissant._src.structure_graph.nodes.source import FileProperty


@dataclasses.dataclass(frozen=True, repr=False)
class Concatenate(Operation):
    """Concatenates several pd.DataFrames into one."""

    node: FileSet

    def call(self, *args: list[Path]) -> pd.DataFrame:
        """See class' docstring."""
        assert len(args) > 0, "No dataframe to merge."
        files = [file for files in args for file in files]
        return pd.DataFrame({
            FileProperty.filepath: [os.fspath(file.filepath) for file in files],
            FileProperty.filename: [file.filename for file in files],
            FileProperty.fullpath: [os.fspath(file.fullpath) for file in files],
        })



from mlcroissant._src.operation_graph.operations.concatenate import Concatenate
from mlcroissant._src.operation_graph.operations.data import Data
from mlcroissant._src.operation_graph.operations.download import Download
from mlcroissant._src.operation_graph.operations.extract import Extract
from mlcroissant._src.operation_graph.operations.field import ReadFields
from mlcroissant._src.operation_graph.operations.filter import FilterFiles
from mlcroissant._src.operation_graph.operations.init import InitOperation
from mlcroissant._src.operation_graph.operations.join import Join
from mlcroissant._src.operation_graph.operations.local_directory import LocalDirectory
from mlcroissant._src.operation_graph.operations.read import Read

__all__ = [
    "Concatenate",
    "Data",
    "Download",
    "Extract",
    "FilterFiles",
    "InitOperation",
    "Join",
    "LocalDirectory",
    "Read",
    "ReadFields",
]



"""PyTorch dataloader-based public API."""

from mlcroissant._src.torch.torch_adapter.dataloader import LoaderFactory
from mlcroissant._src.torch.torch_adapter.dataloader import LoaderSpecificationDataType

__all__ = ["LoaderFactory", "LoaderSpecificationDataType"]



"""LocalDirectory operation module."""

import dataclasses

from etils import epath

from mlcroissant._src.core.path import Path
from mlcroissant._src.operation_graph.base_operation import Operation
from mlcroissant._src.structure_graph.nodes.file_set import FileSet


@dataclasses.dataclass(frozen=True, repr=False)
class LocalDirectory(Operation):
    """Defines a local directory to read files from."""

    node: FileSet
    folder: epath.Path

    def call(self, *args):
        """See class' docstring."""
        del args  # unused
        return Path(
            filepath=self.folder,
            fullpath=self.folder,
        )



"""Lazy imports for optional dependencies."""

from __future__ import annotations

import importlib
import types


def _try_import(module_name: str, package_name: str | None = None):
    """Tries importing a module, with an informative error message on failure.

    Args:
        module_name: The name of the module in Python, e.g.: "import git".
        package_name: The name of the package on PyPI, e.g.: "pip install GitPython".
    """
    if package_name is None:
        package_name = module_name
    try:
        return importlib.import_module(module_name)
    except ImportError as exception:
        if package_name is None:
            package_name = module_name
        installs = f"`pip install {package_name}`"
        error = (
            f"Failed importing {module_name}. This likely means that the dataset"
            " requires additional dependencies that have to be manually installed"
            f" (usually with {installs}). See the optional dependencies listed in"
            " pyproject.toml."
        ).format(name=module_name)
        raise ModuleNotFoundError(error) from exception


class cached_class_property(classmethod):
    """Cached class property decorator.

    Equivalent of @classmethod + @functools.cached_property.
    """

    def __init__(self, func):
        """Constructor with cache."""
        self._func = func
        self._cache = {}

    def __get__(self, obj, objtype=None):
        """Cached getter."""
        if objtype not in self._cache:
            self._cache[objtype] = self._func(objtype)
        return self._cache[objtype]


class OptionalDependencies(object):
    """Optional dependencies can be heavy and need not be loaded by all clients.

    Some datasets require optional dependencies for data generation. To allow for
    the default installation to remain lean, those heavy dependencies are
    lazily imported here.

    Warning: we haven't found a way to properly type the return of each property.

    Usage:

    - Add a new optional dependency below:

    ```python
    @cached_class_property
    def foo(cls):
        return _try_import("goo")
    ```

    - Use this new dependency:

    ```python
    from mlcroissant._src.core.optional import deps

    foo = deps.foo  # use the `foo` module here in your code
    ```
    """

    @cached_class_property
    def git(cls) -> types.ModuleType:
        """Cached git module."""
        return _try_import("git", package_name="GitPython")

    @cached_class_property
    def PIL_Image(cls) -> types.ModuleType:  # pylint: disable=invalid-name
        """Cached PIL module."""
        return _try_import("PIL.Image", package_name="Pillow")

    @cached_class_property
    def librosa(cls) -> types.ModuleType:  # pylint: disable=invalid-name
        """Cached librosa module."""
        return _try_import("librosa", package_name="librosa")

    @cached_class_property
    def torchdata_datapipes(cls) -> types.ModuleType:
        """Cached torchdata module."""
        return _try_import("torchdata.datapipes", package_name="torchdata")


deps = OptionalDependencies



"""bounding_box_test module."""

import pytest

from mlcroissant._src.core.ml import bounding_box


def test_parse():
    assert bounding_box.parse([1, 2, 3, 4]) == [1.0, 2.0, 3.0, 4.0]
    assert bounding_box.parse("1 2 3 4") == [1.0, 2.0, 3.0, 4.0]
    assert bounding_box.parse("1.0 2 3.0 4.0") == [1.0, 2.0, 3.0, 4.0]
    with pytest.raises(ValueError, match="Wrong format"):
        bounding_box.parse(42)
    with pytest.raises(ValueError, match="should have a length of"):
        bounding_box.parse([1, 2, 3, 4, 5])
    with pytest.raises(ValueError, match="can be converted to floats"):
        bounding_box.parse(["one", "two", "three", "four"])



"""Read operation module."""

import dataclasses
import enum
import gzip
import json
import pathlib

from etils import epath
import pandas as pd

from mlcroissant._src.core.constants import EncodingFormat
from mlcroissant._src.core.git import download_git_lfs_file
from mlcroissant._src.core.git import is_git_lfs_file
from mlcroissant._src.core.path import Path
from mlcroissant._src.operation_graph.base_operation import Operation
from mlcroissant._src.operation_graph.operations.download import is_url
from mlcroissant._src.operation_graph.operations.parse_json import parse_json_content
from mlcroissant._src.structure_graph.nodes.field import Field
from mlcroissant._src.structure_graph.nodes.file_object import FileObject
from mlcroissant._src.structure_graph.nodes.file_set import FileSet
from mlcroissant._src.structure_graph.nodes.source import FileProperty


class ReadingMethod(enum.Enum):
    """Reading method derived from the fields that consume the FileObject/FileSet."""

    CONTENT = enum.auto()
    JSON = enum.auto()
    LINES = enum.auto()
    NONE = enum.auto()


def _reading_method(
    node: FileObject | FileSet, fields: tuple[Field, ...]
) -> ReadingMethod:
    """Extracts the reading method from the fields.

    If several reading methods are found, we raise an error for now. Indeed, it is
    unlikely that the same FileObject/FileSet has to be read in different manners. Also,
    an alternative solution is to define n FileObjects/FileSets when you have n
    different reading methods.
    """
    reading_methods: set[ReadingMethod] = set()
    for field in fields:
        extract = field.source.extract
        if extract.column:
            reading_methods.add(ReadingMethod.CONTENT)
        elif extract.file_property == FileProperty.lines:
            reading_methods.add(ReadingMethod.LINES)
        elif extract.file_property == FileProperty.content:
            reading_methods.add(ReadingMethod.CONTENT)
        elif extract.json_path:
            reading_methods.add(ReadingMethod.JSON)
    if len(reading_methods) == 0:
        return ReadingMethod.NONE
    if len(reading_methods) > 1:
        raise ValueError(
            f"Cannot read {node=}. The fields use several reading methods:"
            f" {reading_methods}. Reading the same FileObject/FileSet using different"
            " reading methods has yet to be implemented. Please, create an issue"
            " (https://github.com/mlcommons/croissant/issues/new) if your dataset"
            " requires this feature. Alternatively, you can use two different"
            " FileObject/FileSet pointing to the same resource."
        )
    return next(iter(reading_methods))


def _should_append_line_numbers(fields: tuple[Field, ...]) -> bool:
    """Checks whether at least one field requires listing the line numbers."""
    for field in fields:
        if field.source.extract.file_property == FileProperty.lineNumbers:
            return True
    return False


@dataclasses.dataclass(frozen=True, repr=False)
class Read(Operation):
    """Reads from a file and output a pd.DataFrame."""

    node: FileObject | FileSet
    folder: epath.Path
    fields: tuple[Field, ...]

    def _read_file_content(self, encoding_format: str, file: Path) -> pd.DataFrame:
        """Extracts the `source` file to `target`."""
        filepath = file.filepath
        if is_git_lfs_file(filepath):
            download_git_lfs_file(file)
        reading_method = _reading_method(self.node, self.fields)

        with filepath.open("rb") as file:
            # TODO(https://github.com/mlcommons/croissant/issues/635).
            if filepath.suffix == ".gz":
                file = gzip.open(file, "rt", newline="")
            if encoding_format == EncodingFormat.CSV:
                return pd.read_csv(file)
            elif encoding_format == EncodingFormat.TSV:
                return pd.read_csv(file, sep="\t")
            elif encoding_format == EncodingFormat.JSON:
                json_content = json.load(file)
                if reading_method == ReadingMethod.JSON:
                    return parse_json_content(json_content, self.fields)
                else:
                    # Raw files are returned as a one-line pd.DataFrame.
                    return pd.DataFrame({
                        FileProperty.content: [json_content],
                    })
            elif encoding_format == EncodingFormat.JSON_LINES:
                return pd.read_json(file, lines=True)
            elif encoding_format == EncodingFormat.PARQUET:
                try:
                    df = pd.read_parquet(file)
                    # Sometimes the author already set an index in Parquet, so we want
                    # to reset it to always have the same format.
                    df.reset_index(inplace=True)
                    return df
                except ImportError as e:
                    raise ImportError(
                        "Missing dependency to read Parquet files. pyarrow is not"
                        " installed. Please, install `pip install"
                        " mlcroissant[parquet]`."
                    ) from e
            elif encoding_format == EncodingFormat.TEXT:
                if reading_method == ReadingMethod.LINES:
                    return pd.read_csv(
                        filepath, header=None, names=[FileProperty.lines]
                    )
                else:
                    return pd.DataFrame({
                        FileProperty.content: [file.read()],
                    })
            elif (
                encoding_format == EncodingFormat.MP3
                or encoding_format == EncodingFormat.JPG
            ):
                return pd.DataFrame({
                    FileProperty.content: [file.read()],
                })
            else:
                raise ValueError(
                    f"Unsupported encoding format for file: {encoding_format}"
                )

    def call(self, files: list[Path] | Path) -> pd.DataFrame:
        """See class' docstring."""
        if isinstance(files, Path):
            files = [files]
        file_contents = []
        for file in files:
            # The FileObject is extracted from another FileObject/FileSet:
            if (
                isinstance(self.node, FileObject)
                and self.node.content_url
                and self.node.contained_in
            ):
                content_url = self.node.content_url
                file = Path(
                    filepath=file.filepath / content_url,
                    fullpath=pathlib.PurePath(content_url),
                )
            # The FileObject comes from disk:
            elif (
                isinstance(self.node, FileObject)
                and self.node.content_url
                and not is_url(self.node.content_url)
            ):
                # Read from the local path
                assert file.filepath.exists(), (
                    f'In node "{self.node.uuid}", file "{self.node.content_url}" is'
                    " either an invalid URL or an invalid path."
                )
            assert self.node.encoding_format, "Encoding format is not specified."
            file_content = self._read_file_content(self.node.encoding_format, file)
            if _should_append_line_numbers(self.fields):
                file_content[FileProperty.lineNumbers] = range(len(file_content))
            file_content[FileProperty.filepath] = file.filepath
            file_content[FileProperty.filename] = file.filename
            file_content[FileProperty.fullpath] = file.fullpath
            file_contents.append(file_content)
        return pd.concat(file_contents)



"""Filter operation module."""

import dataclasses
import fnmatch
import functools
import os
import pathlib
import re

from etils import epath

from mlcroissant._src.core.path import get_fullpath
from mlcroissant._src.core.path import Path
from mlcroissant._src.operation_graph.base_operation import Operation
from mlcroissant._src.structure_graph.nodes.file_set import FileSet


@functools.cache
def _memoized_regex(pattern: str) -> re.Pattern:
    regex = fnmatch.translate(pattern)
    return re.compile(regex)


def match_path(patterns: list[str] | None, path: pathlib.PurePath) -> bool:
    """Returns True if at least one pattern matches path."""
    if not patterns:
        return True
    for pattern in patterns:
        regex = _memoized_regex(pattern)
        if regex.match(os.fspath(path)):
            return True
    return False


@dataclasses.dataclass(frozen=True, repr=False)
class FilterFiles(Operation):
    """Filters files within a FileSet given a glob pattern."""

    node: FileSet

    def call(self, *paths: Path) -> list[Path]:
        """See class' docstring."""
        if self.node.includes is None:
            raise ValueError("cannot filter files without `includes`.")
        included_files: list[Path] = []
        for path in paths:
            for basepath, _, files in path.filepath.walk():  # type: ignore  # https://github.com/python/mypy/issues/11880
                for file in files:
                    filepath = epath.Path(basepath) / file
                    fullpath = get_fullpath(filepath, path.filepath)
                    match_includes = match_path(self.node.includes, fullpath)
                    if match_includes:
                        included_files.append(
                            Path(
                                filepath=filepath,
                                fullpath=fullpath,
                            )
                        )
        # We need to sort `files` to have a deterministic/reproducible order.
        included_files.sort()
        return sorted(included_files)



"""field_test module."""

import tempfile
from unittest import mock

from etils import epath
import numpy as np
import pandas as pd
from PIL import Image
import pytest

from mlcroissant._src.core.constants import DataType
from mlcroissant._src.core.context import Context
from mlcroissant._src.core.context import CroissantVersion
from mlcroissant._src.operation_graph.base_operation import Operations
from mlcroissant._src.operation_graph.operations import field
from mlcroissant._src.operation_graph.operations import ReadFields
from mlcroissant._src.structure_graph.nodes.field import Field
from mlcroissant._src.structure_graph.nodes.file_object import FileObject
from mlcroissant._src.structure_graph.nodes.metadata import Metadata
from mlcroissant._src.structure_graph.nodes.record_set import RecordSet
from mlcroissant._src.structure_graph.nodes.source import Extract
from mlcroissant._src.structure_graph.nodes.source import FileProperty
from mlcroissant._src.structure_graph.nodes.source import Source
from mlcroissant._src.structure_graph.nodes.source import Transform
from mlcroissant._src.tests.nodes import empty_record_set
from mlcroissant._src.tests.operations import operations
from mlcroissant._src.tests.versions import parametrize_conforms_to


def test_str_representation():
    operation = field.ReadFields(operations=operations(), node=empty_record_set)
    assert str(operation) == "ReadFields(record_set_name)"


@parametrize_conforms_to()
@pytest.mark.parametrize(
    ["value", "data_type", "expected"],
    [
        [b"iambytes", bytes, b"iambytes"],
        ["iamstring", bytes, b"iamstring"],
        [8, bytes, b"8"],
        [1, float, 1.0],
        ["1", float, 1.0],
        [1.0, float, 1.0],
        ["2024-12-10", pd.Timestamp, pd.Timestamp("2024-12-10")],
    ],
)
def test_cast_value(conforms_to, value, data_type, expected):
    ctx = Context(conforms_to=conforms_to)
    assert field._cast_value(ctx, value, data_type) == expected


@parametrize_conforms_to()
@pytest.mark.parametrize(
    ["value", "data_type"],
    [
        [np.nan, bool],
        [np.nan, bytes],
        [np.nan, pd.Timestamp],
        [np.nan, float],
        [np.nan, int],
    ],
)
def test_cast_value_nan(conforms_to, value, data_type):
    ctx = Context(conforms_to=conforms_to)
    assert np.isnan(field._cast_value(ctx, value, data_type))


@parametrize_conforms_to()
@mock.patch.object(Image, "open", return_value="opened_image")
def test_cast_value_image(open_mock, conforms_to):
    ctx = Context(conforms_to=conforms_to)
    expected = field._cast_value(ctx, b"PNG...Some image...", DataType.IMAGE_OBJECT)
    open_mock.assert_called_once()
    assert expected == "opened_image"


def test_readfield_with_subfields():
    with tempfile.TemporaryDirectory() as tempdir:
        csv_file = epath.Path(tempdir, "dummy_csv.csv")
        with csv_file.open("w") as f:
            f.write("latitude,longitude,names,surnames\n")
            f.write("1,1,Anna-Maria,Rossi-Bianchi\n")
            f.write("2,2,Giulia,Ferrari\n")
            f.write("1,3,,\n")
        # Nodes to define metadata.
        distribution = [
            FileObject(
                name="file",
                id="file_id",
                content_url=csv_file,
                sha256="None",
                encoding_format="text/csv",
            )
        ]
        fields = [
            # One field with subfields.
            Field(
                name="coordinates",
                id="main/coordinates",
                data_types=[DataType.TEXT],
                sub_fields=[
                    Field(
                        name="latitude",
                        id="main/coordinates/latitude",
                        data_types=[DataType.INTEGER],
                        source=Source(
                            file_object="file_id",
                            extract=Extract(column="latitude"),
                        ),
                    ),
                    Field(
                        name="longitude",
                        id="main/coordinates/longitude",
                        data_types=[DataType.INTEGER],
                        source=Source(
                            file_object="file_id",
                            extract=Extract(column="longitude"),
                        ),
                    ),
                ],
            ),
            # One field with repeated subfields.
            Field(
                name="checked_users",
                id="main/checked_users",
                data_types=[DataType.TEXT],
                repeated=True,
                sub_fields=[
                    Field(
                        name="name",
                        id="main/checked_users/name",
                        data_types=[DataType.TEXT],
                        source=Source(
                            file_object="file_id",
                            extract=Extract(column="names"),
                            transforms=[Transform(separator="-")],
                        ),
                    ),
                    Field(
                        name="surname",
                        id="main/checked_users/surname",
                        data_types=[DataType.TEXT],
                        source=Source(
                            file_object="file_id",
                            extract=Extract(column="surnames"),
                            transforms=[Transform(separator="-")],
                        ),
                    ),
                ],
            ),
        ]
        with mock.patch.object(RecordSet, "check_joins_in_fields") as mock_check_joins:
            mock_check_joins.return_value = True
            record_set = RecordSet(name="main", id="main", fields=fields)
            record_sets = [record_set]
            ctx = Context(conforms_to=CroissantVersion.V_1_0)
            Metadata(
                ctx=ctx,
                name="metadata",
                url="url.com",
                distribution=distribution,
                record_sets=record_sets,
            )
            read_field = ReadFields(operations=Operations(), node=record_sets[0])
            df = pd.read_csv(
                csv_file
            )  # pd.DataFrame({FileProperty.filepath: [csv_file]})
            expected = [
                {
                    "main/coordinates": {
                        "main/coordinates/latitude": 1,
                        "main/coordinates/longitude": 1,
                    },
                    "main/checked_users": [
                        {
                            "main/checked_users/name": b"Anna",
                            "main/checked_users/surname": b"Rossi",
                        },
                        {
                            "main/checked_users/name": b"Maria",
                            "main/checked_users/surname": b"Bianchi",
                        },
                    ],
                },
                {
                    "main/coordinates": {
                        "main/coordinates/latitude": 2,
                        "main/coordinates/longitude": 2,
                    },
                    "main/checked_users": [
                        {
                            "main/checked_users/name": b"Giulia",
                            "main/checked_users/surname": b"Ferrari",
                        },
                    ],
                },
                {
                    "main/coordinates": {
                        "main/coordinates/latitude": 1,
                        "main/coordinates/longitude": 3,
                    },
                    "main/checked_users": [
                        {
                            "main/checked_users/name": None,
                            "main/checked_users/surname": None,
                        },
                    ],
                },
            ]
            result = list(read_field.call(df))
            assert result == expected


@pytest.mark.parametrize(
    "separator",
    [
        b"\n",
        b"\r",
        b"\r\n",
    ],
)
def test_extract_lines(separator):
    with tempfile.TemporaryDirectory() as tempdir:
        # Create the underlying file.
        content = (
            b"bon jour  "
            + separator
            + separator
            + b" h\xc3\xa9llo "
            + separator
            + b"hallo "
            + separator
        )
        path = tempdir + "/file.txt"
        with open(path, "wb") as f:
            f.write(content)

        # Create all needed nodes.
        distribution = [
            FileObject(
                name="file",
                id="file_id",
                content_url=path,
                sha256="None",
                encoding_format="text/plain",
            )
        ]
        fields = []
        fields.append(
            Field(
                name="line",
                id="main/line",
                data_types=[DataType.TEXT],
                source=Source(
                    file_object="file_id",
                    extract=Extract(file_property=FileProperty.lines),
                ),
            )
        )
        fields.append(
            Field(
                name="line_number",
                id="main/line_number",
                data_types=[DataType.INTEGER],
                source=Source(
                    file_object="file_id",
                    extract=Extract(file_property=FileProperty.lineNumbers),
                ),
            )
        )
        fields.append(
            Field(
                name="filename",
                id="main/filename",
                data_types=[DataType.TEXT],
                source=Source(
                    file_object="file_id",
                    extract=Extract(file_property=FileProperty.filepath),
                    transforms=[Transform(regex=".*\\/(\\w*)\\.txt")],
                ),
            )
        )
        with mock.patch.object(RecordSet, "check_joins_in_fields") as mock_check_joins:
            mock_check_joins.return_value = True
            record_set = RecordSet(name="main", id="main", fields=fields)
            record_sets = [record_set]
            ctx = Context(conforms_to=CroissantVersion.V_1_0)
            Metadata(
                ctx=ctx,
                name="metadata",
                url="url.com",
                distribution=distribution,
                record_sets=record_sets,
            )
            read_field = ReadFields(operations=Operations(), node=record_sets[0])
            df = pd.DataFrame({FileProperty.filepath: [path]})
            expected = [
                {
                    "main/line_number": 0,
                    "main/line": b"bon jour  ",
                    "main/filename": b"file",
                },
                {"main/line_number": 1, "main/line": b"", "main/filename": b"file"},
                {
                    "main/line_number": 2,
                    "main/line": b" h\xc3\xa9llo ",
                    "main/filename": b"file",
                },
                {
                    "main/line_number": 3,
                    "main/line": b"hallo ",
                    "main/filename": b"file",
                },
            ]
            result = list(read_field.call(df))
            assert result == expected


@pytest.mark.parametrize(
    ["value", "source", "data_type", "expected_value", "repeated"],
    [
        # Capturing group
        [
            "train1234",
            Source(transforms=[Transform(regex="(train|val)\\d\\d\\d\\d")]),
            DataType.TEXT,
            "train",
            False,
        ],
        [
            ["train1234", "train5678", "val1111"],
            Source(transforms=[Transform(regex="(train|val)\\d\\d\\d\\d")]),
            DataType.TEXT,
            ["train", "train", "val"],
            True,
        ],
        [
            epath.Path("path/to/train1234"),
            Source(transforms=[Transform(regex=".*/(train|val)\\d\\d\\d\\d")]),
            DataType.TEXT,
            "train",
            False,
        ],
        # Non working capturing group
        [
            "foo1234",
            Source(transforms=[Transform(regex="(train|val)\\d\\d\\d\\d")]),
            DataType.TEXT,
            "foo1234",
            False,
        ],
        [
            {"one": {"two": "expected_value"}, "three": "non_expected_value"},
            Source(transforms=[Transform(json_path="one.two")]),
            DataType.TEXT,
            "expected_value",
            False,
        ],
        [
            pd.Timestamp("2024-12-10 12:00:00"),
            Source(transforms=[Transform(format="%Y-%m-%d")]),
            DataType.DATE,
            "2024-12-10",
            False,
        ],
        [
            "2024-12-10 12:00:00",
            Source(transforms=[Transform(format="%Y-%m-%d")]),
            DataType.DATE,
            "2024-12-10",
            False,
        ],
    ],
)
def test_apply_transforms_fn(value, source, data_type, expected_value, repeated):
    f = Field(id="test", name="test", data_types=data_type, source=source)
    assert field.apply_transforms_fn(value, f, repeated=repeated) == expected_value



"""graph_test module."""

import json

from etils import epath
from rdflib import term

from mlcroissant._src.core.context import Context
from mlcroissant._src.structure_graph.nodes.metadata import Metadata

Literal = term.Literal


# If this test fails, you probably manually updated a dataset in datasets/.
# Please, use scripts/migrations/migrate.py to migrate datasets.
def jsonld_to_python_to_jsonld(path):
    with path.open() as f:
        json_ld = json.load(f)
    ctx = Context()
    metadata = Metadata.from_file(ctx, {})
    result = metadata.to_json()
    # `distribution` may not be in the right order:
    if "distribution" in result:
        distribution = result.pop("distribution")
        distribution = {file["name"]: file for file in distribution}
        expected_distribution = json_ld.pop("distribution")
        expected_distribution = {file["name"]: file for file in expected_distribution}
        assert distribution == expected_distribution
    # Check the expected JSON-LD:
    assert result == json_ld
    assert not ctx.issues.errors


def test_jsonld_to_python_to_jsonld():
    dataset_folder = (
        epath.Path(__file__).parent.parent.parent.parent.parent.parent / "datasets"
    )
    json_ld_paths = [path for path in dataset_folder.glob("*/*.json")]
    for path in json_ld_paths:
        jsonld_to_python_to_jsonld(path)



"""issues module."""

import dataclasses
from typing import Any

WRONG_ID_MSG = (
    "Note that `@id`s should be valid URIs. If you wish to use reserved characters in"
    " your `@id`s, you should URL-encode them."
)


class ValidationError(Exception):
    """Error during the validation of the format."""


class GenerationError(Exception):
    """Error during the generation of the dataset."""


class WarningException(Exception):
    """Exception triggering a warning."""


class ErrorException(Exception):
    """Exception triggering an error."""


@dataclasses.dataclass(frozen=True)
class Issues:
    """Issues during the validation of the format.

    Issues can either be errors (blocking) or warnings (informative).

    We use sets to represent errors and warnings to avoid repeated strings.
    """

    _errors: set[tuple[str, Any]] = dataclasses.field(default_factory=set, hash=False)
    _warnings: set[tuple[str, Any]] = dataclasses.field(default_factory=set, hash=False)

    def _wrap_in_context(self, context: str | None, issue: str) -> str:
        if context is None:
            return issue
        return f"[{context}] {issue}"

    def add_error(self, error: str, node: Any = None):
        """Mutates self.errors with a new error."""
        self._errors.add((error, node))

    def add_warning(self, warning: str, node: Any = None):
        """Mutates self.warnings with a new warning."""
        self._warnings.add((warning, node))

    @property
    def errors(self) -> set[str]:
        """Outputs errors as a set of human-readable strings."""
        errors = set()
        for error, node in self._errors:
            if get_issue_context := getattr(node, "get_issue_context", None):
                if callable(get_issue_context):
                    errors.add(f"[{get_issue_context()}] {error}")
                    continue
            errors.add(error)
        return errors

    @property
    def warnings(self) -> set[str]:
        """Outputs warnings as a set of human-readable strings."""
        warnings = set()
        for error, node in self._warnings:
            if get_issue_context := getattr(node, "get_issue_context", None):
                if callable(get_issue_context):
                    warnings.add(f"[{get_issue_context()}] {error}")
                    continue
            warnings.add(error)
        return warnings

    def report(self) -> str:
        """Reports errors and warnings in a string."""
        message = ""
        # Sort before printing because sets are not ordered.
        for issues, issue_type in [
            (sorted(self.errors), "error(s)"),
            (sorted(self.warnings), "warning(s)"),
        ]:
            num_issues = len(issues)
            if num_issues:
                message += (
                    f"Found the following {len(issues)} {issue_type} during the"
                    " validation:\n"
                )
                for issue in issues:
                    message += f"  -  {issue}\n"
        return message.strip()



"""extract_test module."""

import pathlib

from etils import epath

from mlcroissant._src.core.path import get_fullpaths
from mlcroissant._src.operation_graph.operations.extract import Extract
from mlcroissant._src.tests.nodes import empty_file_object
from mlcroissant._src.tests.operations import operations


def test_str_representation():
    operation = Extract(operations=operations(), node=empty_file_object)
    assert str(operation) == "Extract(file_object_name)"


def test_get_fullpaths():
    files = [
        epath.Path("/path/to/extract/file1"),
        epath.Path("/path/to/extract/file2/foo"),
        epath.Path("/path/to/extract/file2/bar"),
    ]
    extract_dir = epath.Path("/path/to/extract")
    assert get_fullpaths(files, extract_dir) == [
        pathlib.PurePath("file1"),
        pathlib.PurePath("file2/foo"),
        pathlib.PurePath("file2/bar"),
    ]



"""Versions module."""

import re
from typing import Any

from mlcroissant._src.core.issues import ErrorException
from mlcroissant._src.core.issues import WarningException

# Adapted from https://pypi.org/project/semver.
VERSION_PATTERN_REGEX = re.compile(
    r"""
        ^
        (?P<major>0|[1-9]\d*)
        (
          \.
          (?P<minor>0|[1-9]\d*)
          (
            \.
            (?P<patch>0|[1-9]\d*)
            (?:-(?P<prerelease>
                (?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)
                (?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*
            ))?
            (?:\+(?P<build>
                [0-9a-zA-Z-]+
                (?:\.[0-9a-zA-Z-]+)*
            ))?
          )?
        )?
        $
    """,
    re.VERBOSE,
)


def cast_version(version: Any) -> str | None:
    """Casts the dataset version and returns a normalized string version.

    A valid version follows Semantic Versioning 2.0.0 `MAJOR.MINOR.PATCH`.
    For more information: https://semver.org/spec/v2.0.0.html.
    """
    # Version is a recommended but not mandatory attribute.
    if version is None:
        return None

    if isinstance(version, str):
        if not VERSION_PATTERN_REGEX.match(version):
            raise WarningException(
                f"Version doesn't follow MAJOR.MINOR.PATCH: {version}. For more"
                " information refer to: https://semver.org/spec/v2.0.0.html"
            )
        return version
    elif isinstance(version, int):
        return f"{version}.0.0"
    elif isinstance(version, float):
        return f"{version}.0"
    raise ErrorException(
        f"The version should be a string or a number. Got: {type(version)}."
    )



"""records_test module."""

import pandas as pd

from mlcroissant._src.tests.records import record_to_python


def test_record_to_python():
    assert record_to_python({
        "key1": 1,
        "key2": {
            "key3": pd.Timestamp("2017-01-01T12"),
            "key4": {"key5": b"foo", "key6": float("nan")},
        },
    }) == {
        "key1": 1,
        "key2": {"key3": "2017-01-01 12:00:00", "key4": {"key5": "foo", "key6": None}},
    }
    assert record_to_python({
        "image": (
            "<PIL.PngImagePlugin.PngImageFile image mode=L size=28x28 at"
            " 0x7F40B2EB5420>"
        )
    }) == {
        "image": (
            "<PIL.PngImagePlugin.PngImageFile image mode=L size=28x28 at"
            " <MEMORY_ADDRESS>>"
        )
    }



"""constants module."""

# flake8: noqa

import os

from etils import epath
import rdflib
from rdflib import namespace
from rdflib import term

# MLCommons-defined URIs.
ML_COMMONS_V_0_8 = rdflib.Namespace("http://mlcommons.org/schema/")
ML_COMMONS_V_1_0 = rdflib.Namespace("http://mlcommons.org/croissant/")


# Value for the base IRI used when expanding @ids in the JSON-LD.
BASE_IRI = "cr_base_iri/"


# ctx: Context is untyped to avoid cyclic dependencies. A unit test tests the behaviour.
def ML_COMMONS(ctx) -> rdflib.Namespace:
    """Switches the main Croissant namespace according to the version."""
    if ctx.is_v0():  # pytype: disable=attribute-error
        return ML_COMMONS_V_0_8
    else:
        return ML_COMMONS_V_1_0


ML_COMMONS_ARRAY_SHAPE = lambda ctx: ML_COMMONS(ctx).arrayShape
ML_COMMONS_CITE_AS = lambda ctx: (
    SCHEMA_ORG_CITATION if ctx.is_v0() else ML_COMMONS(ctx).citeAs
)
ML_COMMONS_COLUMN = lambda ctx: ML_COMMONS(ctx).column
ML_COMMONS_DATA = lambda ctx: ML_COMMONS(ctx).data
ML_COMMONS_DATA_BIASES = lambda ctx: ML_COMMONS(ctx).dataBiases
ML_COMMONS_DATA_COLLECTION = lambda ctx: ML_COMMONS(ctx).dataCollection
ML_COMMONS_DATA_TYPE = lambda ctx: ML_COMMONS(ctx).dataType
ML_COMMONS_DATA_TYPE_BOUNDING_BOX = lambda ctx: ML_COMMONS(ctx).BoundingBox
ML_COMMONS_EQUIVALENT_PROPERTY = lambda ctx: ML_COMMONS(ctx).equivalentProperty
ML_COMMONS_EXAMPLES = lambda ctx: ML_COMMONS(ctx).examples
ML_COMMONS_EXCLUDES = lambda ctx: ML_COMMONS(ctx).excludes
ML_COMMONS_EXTRACT = lambda ctx: ML_COMMONS(ctx).extract
ML_COMMONS_FILE_PROPERTY = lambda ctx: ML_COMMONS(ctx).fileProperty
ML_COMMONS_FIELD = lambda ctx: ML_COMMONS(ctx).field
ML_COMMONS_FIELD_TYPE = lambda ctx: ML_COMMONS(ctx).Field
ML_COMMONS_FILE_OBJECT = lambda ctx: ML_COMMONS(ctx).fileObject
ML_COMMONS_FILE_SET = lambda ctx: ML_COMMONS(ctx).fileSet
# ML_COMMONS.format is understood as the `format` method on the class Namespace.
ML_COMMONS_FORMAT = lambda ctx: ML_COMMONS(ctx)["format"]
ML_COMMONS_INCLUDES = lambda ctx: ML_COMMONS(ctx).includes
ML_COMMONS_IS_ARRAY = lambda ctx: ML_COMMONS(ctx).isArray
ML_COMMONS_IS_ENUMERATION = lambda ctx: ML_COMMONS(ctx).isEnumeration
ML_COMMONS_IS_LIVE_DATASET = lambda ctx: ML_COMMONS(ctx).isLiveDataset
ML_COMMONS_JSON_PATH = lambda ctx: ML_COMMONS(ctx).jsonPath
ML_COMMONS_PARENT_FIELD = lambda ctx: ML_COMMONS(ctx).parentField
ML_COMMONS_PATH = lambda ctx: ML_COMMONS(ctx).path
ML_COMMONS_PERSONAL_SENSITVE_INFORMATION = lambda ctx: ML_COMMONS(
    ctx
).personalSensitiveInformation
ML_COMMONS_RECORD_SET = lambda ctx: ML_COMMONS(ctx).recordSet
ML_COMMONS_RECORD_SET_TYPE = lambda ctx: ML_COMMONS(ctx).RecordSet
ML_COMMONS_REFERENCES = lambda ctx: ML_COMMONS(ctx).references
ML_COMMONS_REGEX = lambda ctx: ML_COMMONS(ctx).regex
ML_COMMONS_REPEATED = lambda ctx: ML_COMMONS(ctx).repeated
# ML_COMMONS.replace is understood as the `replace` method on the class Namespace.
ML_COMMONS_REPLACE = lambda ctx: ML_COMMONS(ctx)["replace"]
ML_COMMONS_SEPARATOR = lambda ctx: ML_COMMONS(ctx).separator
ML_COMMONS_SOURCE = lambda ctx: ML_COMMONS(ctx).source
ML_COMMONS_SUB_FIELD = lambda ctx: ML_COMMONS(ctx).subField
ML_COMMONS_SUB_FIELD_TYPE = lambda ctx: ML_COMMONS(ctx).SubField
ML_COMMONS_TRANSFORM = lambda ctx: ML_COMMONS(ctx).transform

# Croissant RAI extension
# V1.0 namespace
RAI = rdflib.Namespace("http://mlcommons.org/croissant/RAI/")
# Attributes
ML_COMMONS_RAI_DATA_COLLECTION = RAI.dataCollection
ML_COMMONS_RAI_DATA_COLLECTION_TYPE = RAI.dataCollectionType
ML_COMMONS_RAI_DATA_COLLECTION_MISSING_DATA = RAI.dataCollectionMissingData
ML_COMMONS_RAI_DATA_COLLECTION_RAW_DATA = RAI.dataCollectionRawData
ML_COMMONS_RAI_DATA_COLLECTION_TIME_FRAME = RAI.dataCollectionTimeFrame
ML_COMMONS_RAI_DATA_IMPUTATION_PROTOCOL = RAI.dataImputationProtocol
ML_COMMONS_RAI_DATA_PREPROCESSING_PROTOCOL = RAI.dataPreprocessingProtocol
ML_COMMONS_RAI_DATA_DATA_MANIPULATION_PROTOCOL = RAI.dataDataManipulationProtocol
ML_COMMONS_RAI_DATA_ANNOTATION_PROTOCOL = RAI.dataAnnotationProtocol
ML_COMMONS_RAI_DATA_ANNOTATION_PLATFORM = RAI.dataAnnotationPlatform
ML_COMMONS_RAI_DATA_ANNOTATION_ANALYSIS = RAI.dataAnnotationAnalysis
ML_COMMONS_RAI_ANNOTATIONS_PER_ITEM = RAI.annotationsPerItem
ML_COMMONS_RAI_ANNOTATOR_DEMOGRAPHICS = RAI.annotatorDemographics
ML_COMMONS_RAI_MACHINE_ANNOTATION_TOOLS = RAI.machineAnnotationTools
ML_COMMONS_RAI_DATA_USE_CASES = RAI.dataUseCases
ML_COMMONS_RAI_DATA_BIASES = RAI.dataBiases
ML_COMMONS_RAI_DATA_LIMITATIONS = RAI.dataLimitations
ML_COMMONS_RAI_DATA_SOCIAL_IMPACT = RAI.dataSocialImpact
ML_COMMONS_RAI_PERSONAL_SENSITIVE_INFORMATION = RAI.personalSensitiveInformation
ML_COMMONS_RAI_DATA_RELEASE_MAINTENANCE_PLAN = RAI.dataReleaseMaintenancePlan

# RDF standard URIs.
# For "@type" key:
RDF_TYPE = namespace.RDF.type

# Dublin Core terms standard URIs.
DCTERMS = "http://purl.org/dc/terms/"
DCTERMS_CONFORMS_TO = namespace.DCTERMS.conformsTo

# Wikidata standard URIs.
WIKIDATA = rdflib.Namespace("https://www.wikidata.org/wiki/")

# Schema.org standard URIs.
SCHEMA_ORG_CITATION: term.URIRef = namespace.SDO.citation
SCHEMA_ORG_CONTAINED_IN = namespace.SDO.containedIn
SCHEMA_ORG_CONTENT_SIZE = namespace.SDO.contentSize
SCHEMA_ORG_CONTENT_URL = namespace.SDO.contentUrl
SCHEMA_ORG_CREATOR = namespace.SDO.creator
SCHEMA_ORG_DATE_CREATED = namespace.SDO.dateCreated
SCHEMA_ORG_DATE_MODIFIED = namespace.SDO.dateModified
SCHEMA_ORG_DATE_PUBLISHED = namespace.SDO.datePublished
SCHEMA_ORG_DATASET = namespace.SDO.Dataset
SCHEMA_ORG_DATA_TYPE_AUDIO_OBJECT = namespace.SDO.AudioObject
SCHEMA_ORG_DATA_TYPE_BOOL = namespace.SDO.Boolean
SCHEMA_ORG_DATA_TYPE_DATE = namespace.SDO.Date
SCHEMA_ORG_DATA_TYPE_FLOAT = namespace.SDO.Float
SCHEMA_ORG_DATA_TYPE_IMAGE_OBJECT = namespace.SDO.ImageObject
SCHEMA_ORG_DATA_TYPE_INTEGER = namespace.SDO.Integer
SCHEMA_ORG_DATA_TYPE_TEXT = namespace.SDO.Text
SCHEMA_ORG_DATA_TYPE_URL = namespace.SDO.URL
SCHEMA_ORG_DESCRIPTION = namespace.SDO.description
SCHEMA_ORG_DISTRIBUTION = namespace.SDO.distribution
SCHEMA_ORG_EMAIL = namespace.SDO.email
SCHEMA_ORG_ENCODING_FORMAT = namespace.SDO.encodingFormat
SCHEMA_ORG_KEYWORDS = namespace.SDO.keywords
SCHEMA_ORG_LICENSE = namespace.SDO.license
SCHEMA_ORG_NAME = namespace.SDO.name
SCHEMA_ORG_PUBLISHER = namespace.SDO.publisher
SCHEMA_ORG_SAME_AS = namespace.SDO.sameAs
SCHEMA_ORG_SHA256 = namespace.SDO.sha256
SCHEMA_ORG_URL = namespace.SDO.url
SCHEMA_ORG_VERSION = namespace.SDO.version

# Schema.org URIs that do not exist yet in the standard.
SCHEMA_ORG = rdflib.Namespace("https://schema.org/")
SCHEMA_ORG_KEY = lambda ctx: SCHEMA_ORG.key if ctx.is_v0() else ML_COMMONS(ctx)["key"]
SCHEMA_ORG_FILE_OBJECT = lambda ctx: (
    SCHEMA_ORG.FileObject if ctx.is_v0() else ML_COMMONS(ctx)["FileObject"]
)
SCHEMA_ORG_FILE_SET = lambda ctx: (
    SCHEMA_ORG.FileSet if ctx.is_v0() else ML_COMMONS(ctx)["FileSet"]
)
SCHEMA_ORG_MD5 = lambda ctx: SCHEMA_ORG.md5 if ctx.is_v0() else ML_COMMONS(ctx)["md5"]

TO_CROISSANT = lambda ctx: {
    ML_COMMONS_ARRAY_SHAPE(ctx): "array_shape",
    ML_COMMONS_CITE_AS(ctx): "cite_as",
    ML_COMMONS_COLUMN(ctx): "csv_column",
    ML_COMMONS_DATA_TYPE(ctx): "data_type",
    ML_COMMONS_DATA(ctx): "data",
    ML_COMMONS_EXTRACT(ctx): "extract",
    ML_COMMONS_FIELD(ctx): "field",
    ML_COMMONS_FILE_PROPERTY(ctx): "file_property",
    ML_COMMONS_FORMAT(ctx): "format",
    ML_COMMONS_INCLUDES(ctx): "includes",
    ML_COMMONS_IS_ARRAY(ctx): "is_array",
    ML_COMMONS_IS_LIVE_DATASET(ctx): "is_live_dataset",
    ML_COMMONS_JSON_PATH(ctx): "json_path",
    ML_COMMONS_REFERENCES(ctx): "references",
    ML_COMMONS_REGEX(ctx): "regex",
    ML_COMMONS_REPLACE(ctx): "replace",
    ML_COMMONS_SEPARATOR(ctx): "separator",
    ML_COMMONS_SOURCE(ctx): "source",
    ML_COMMONS_TRANSFORM(ctx): "transforms",
    DCTERMS_CONFORMS_TO: "conforms_to",
    SCHEMA_ORG_CONTAINED_IN: "contained_in",
    SCHEMA_ORG_CONTENT_SIZE: "content_size",
    SCHEMA_ORG_CONTENT_URL: "content_url",
    SCHEMA_ORG_CREATOR: "creators",
    SCHEMA_ORG_DATE_CREATED: "date_created",
    SCHEMA_ORG_DATE_MODIFIED: "date_modified",
    SCHEMA_ORG_DATE_PUBLISHED: "date_published",
    SCHEMA_ORG_DESCRIPTION: "description",
    SCHEMA_ORG_DISTRIBUTION: "distribution",
    SCHEMA_ORG_ENCODING_FORMAT: "encoding_format",
    SCHEMA_ORG_KEYWORDS: "keywords",
    SCHEMA_ORG_LICENSE: "license",
    SCHEMA_ORG_MD5: "md5",
    SCHEMA_ORG_NAME: "name",
    SCHEMA_ORG_PUBLISHER: "publisher",
    SCHEMA_ORG_SAME_AS: "sameAs",
    SCHEMA_ORG_SHA256: "sha256",
    SCHEMA_ORG_URL: "url",
    SCHEMA_ORG_VERSION: "version",
}

FROM_CROISSANT = lambda ctx: {v: k for k, v in TO_CROISSANT(ctx).items()}

# Environment variables
DEFAULT_CROISSANT_CACHE = "~/.cache/croissant"
CROISSANT_CACHE = epath.Path(
    os.getenv("CROISSANT_CACHE", DEFAULT_CROISSANT_CACHE)
).expanduser()
DOWNLOAD_PATH = CROISSANT_CACHE / "download"
EXTRACT_PATH = CROISSANT_CACHE / "extract"
CROISSANT_GIT_USERNAME = "CROISSANT_GIT_USERNAME"
CROISSANT_GIT_PASSWORD = "CROISSANT_GIT_PASSWORD"
CROISSANT_BASIC_AUTH_USERNAME = "CROISSANT_BASIC_AUTH_USERNAME"
CROISSANT_BASIC_AUTH_PASSWORD = "CROISSANT_BASIC_AUTH_PASSWORD"


class EncodingFormat:
    """Supported MIME Types in Croissant.

    We inherit the wrong naming `encodingFormat` from https://schema.org/encodingFormat.
    """

    CSV = "text/csv"
    GIT = "git+https"
    JPG = "image/jpeg"
    JSON = "application/json"
    JSON_LINES = "application/jsonlines"
    MP3 = "audio/mpeg"
    PARQUET = "application/x-parquet"
    TEXT = "text/plain"
    TSV = "text/tab-separated-values"
    TAR = "application/x-tar"
    ZIP = "application/zip"


class DataType:
    """Data types supported by Croissant."""

    AUDIO_OBJECT = namespace.SDO.AudioObject
    BOOL = namespace.SDO.Boolean
    BOUNDING_BOX = ML_COMMONS_V_1_0.BoundingBox
    DATE = namespace.SDO.Date
    FLOAT = namespace.SDO.Float
    FLOAT16 = ML_COMMONS_V_1_0.Float16
    FLOAT32 = ML_COMMONS_V_1_0.Float32
    FLOAT64 = ML_COMMONS_V_1_0.Float64
    IMAGE_OBJECT = namespace.SDO.ImageObject
    INTEGER = namespace.SDO.Integer
    INT8 = ML_COMMONS_V_1_0.Int8
    INT16 = ML_COMMONS_V_1_0.Int16
    INT32 = ML_COMMONS_V_1_0.Int32
    INT64 = ML_COMMONS_V_1_0.Int64
    SPLIT = ML_COMMONS_V_1_0.Split
    TEXT = namespace.SDO.Text
    URL = namespace.SDO.URL



"""uuid_test module."""

from unittest import mock

import pytest

from mlcroissant._src.core.constants import BASE_IRI
from mlcroissant._src.core.context import Context
from mlcroissant._src.core.context import CroissantVersion
import mlcroissant._src.core.uuid as uuid_lib


@pytest.mark.parametrize(
    ["input", "output"],
    [
        [f"aaa{BASE_IRI}bbb", "bbb"],
        [f"aaa{BASE_IRI}b{BASE_IRI}bb", f"b{BASE_IRI}bb"],
        [{"@id": "aaa"}, "aaa"],
        ["aaa", "aaa"],
        [f"{BASE_IRI}aaa", "aaa"],
        [f"aaa{BASE_IRI}", ""],
    ],
)
def test_uuid_from_jsonld(input, output):
    assert uuid_lib.uuid_from_jsonld(input) == output


@mock.patch.object(uuid_lib, "generate_uuid")
def test_uuid_from_jsonld_generate_uuid(mock_uuid_lib_generate_uuid):
    mock_uuid_lib_generate_uuid.return_value = "123456"
    assert uuid_lib.uuid_from_jsonld(None) == "123456"


@pytest.mark.parametrize(
    ["input", "output"], [[f"aaa{BASE_IRI}bbb", "bbb"], ["aaa", "aaa"]]
)
def test_uuid_to_jsonld(input, output):
    assert uuid_lib.uuid_to_jsonld(input) == output


@pytest.mark.parametrize(
    ["conforms_to", "uuid", "output"],
    [
        [CroissantVersion.V_0_8, "example/uuid", "example/uuid"],
        [CroissantVersion.V_1_0, "example/uuid", {"@id": "example/uuid"}],
        [
            CroissantVersion.V_1_0,
            ["example/uuid1", "example/uuid2"],
            [{"@id": "example/uuid1"}, {"@id": "example/uuid2"}],
        ],
        [CroissantVersion.V_1_0, ["example/uuid1"], {"@id": "example/uuid1"}],
    ],
)
def test_formatted_uuid_to_json(conforms_to, uuid, output):
    ctx = Context(conforms_to=conforms_to)
    assert uuid_lib.formatted_uuid_to_json(ctx=ctx, uuid=uuid) == output



"""Utils for graphs."""

import time

import networkx as nx


def pretty_print_graph(graph: nx.Graph):
    """Pretty prints a NetworkX graph.

    Warning: this function is for debugging purposes only.

    Args:
        graph: Any NetworkX graph.
    """
    agraph = nx.nx_agraph.to_agraph(graph)
    agraph.layout(prog="dot")
    temporary_file = f"/tmp/graph_{time.time()}.svg"
    agraph.draw(temporary_file, args="-Gnodesep=0.01 -Gfont_size=1", prog="dot")
    print(f"Generated a graph and saved it in: {temporary_file}")


def print_graph_traversal(graph: nx.Graph):
    """Pretty prints a NetworkX graph.

    Warning: this function is for debugging purposes only.

    Args:
        graph: Any NetworkX graph.
    """
    visited = {}
    print("--- Graph traversal ---")
    for start, end, _ in nx.edge_bfs(graph):
        for node in [start, end]:
            if node.id not in visited:
                print(f"Visited: {node.id}")
                visited[node.id] = True
    print("Done traversing the graph.")



"""Generates the dataset and yields the first example."""

import json
from typing import Any, Mapping

from absl import app
from absl import flags
from absl import logging
from etils import epath

import mlcroissant as mlc
from mlcroissant._src.tests.records import record_to_python

_NUM_MAX_RECORDS = 10


flags.DEFINE_string(
    "jsonld",
    None,
    "JSON-LD to validate (path to the file or URL).",
    required=True,
)

flags.DEFINE_string(
    "file",
    "",
    "[DEPRECATED] Path to the file to validate.",
    required=False,
)

flags.DEFINE_string(
    "record_set",
    None,
    "The name of the record set to generate.",
)

flags.DEFINE_integer(
    "num_records",
    _NUM_MAX_RECORDS,
    "The number of records to generate. Use `-1` to generate the whole dataset.",
)

flags.DEFINE_bool(
    "debug",
    False,
    "Whether to print debug hints.",
)

flags.DEFINE_bool(
    "update_output",
    False,
    "Whether to update the JSONL output test files.",
)

flags.DEFINE_string(
    "mapping",
    None,
    "Mapping filename->filepath as a Python dict[str, str] to handle manual downloads."
    " If `document.csv` is the FileObject and you downloaded it to"
    ' `~/Downloads/document.csv`, you can specify `--mapping \'{"document.csv":'
    ' "~/Downloads/document.csv"}\'`.',
)

flags.DEFINE_string(
    "filters",
    None,
    'Filters to apply to `Dataset.records`. For example, \'{"data/split": "train"}\'',
)

flags.mark_flag_as_required("jsonld")


FLAGS = flags.FLAGS


def main(argv):
    """Main function launched by the script."""
    del argv
    if FLAGS.file:
        logging.warning("--file is deprecated. Please, use --jsonld with a path or URL")
    jsonld = FLAGS.jsonld or FLAGS.file
    record_set = FLAGS.record_set
    num_records = FLAGS.num_records
    debug = FLAGS.debug
    update_output = FLAGS.update_output
    mapping = FLAGS.mapping
    filters = FLAGS.filters
    return load(
        jsonld=jsonld,
        record_set=record_set,
        num_records=num_records,
        debug=debug,
        update_output=update_output,
        mapping=mapping,
        filters=filters,
    )


def load(
    jsonld: str,
    record_set: str | None,
    num_records: int = _NUM_MAX_RECORDS,
    debug: bool = False,
    update_output: bool = False,
    mapping: str | None = None,
    filters: str | None = None,
):
    """Yields data from the `record_set` in the input Croissant file."""
    if not mapping:
        file_mapping: Mapping[str, Any] = {}
    else:
        try:
            file_mapping = json.loads(mapping)
        except json.JSONDecodeError as e:
            raise ValueError("--mapping should be a valid dict[str, str]") from e
    parsed_filters = None
    if filters:
        try:
            parsed_filters = json.loads(filters)
        except json.JSONDecodeError as e:
            raise ValueError("--filters should be a valid dict[str, str]") from e
    dataset = mlc.Dataset(jsonld, debug=debug, mapping=file_mapping)
    if record_set is None:
        record_sets = ", ".join([f"`{rs.id}`" for rs in dataset.metadata.record_sets])
        raise ValueError(f"--record_set flag should have a value in {record_sets}")
    records = dataset.records(record_set, filters=parsed_filters)
    generate_all_records = num_records == -1
    if generate_all_records:
        print(f"Generating all records from {jsonld}.")
    else:
        print(f"Generating the first {num_records} records from {jsonld}.")
    output_records = []
    for i, record in enumerate(records):
        if not generate_all_records and i >= num_records:
            break
        print(record)
        output_records.append(record_to_python(record))
    print("Done.")
    if update_output and not jsonld.startswith("http"):
        output_folder = epath.Path(jsonld).parent / "output"
        if not output_folder.exists():
            output_folder.mkdir()
        output_file = output_folder / f"{record_set}.jsonl"
        with output_file.open("w") as f:
            for output_record in output_records:
                output_record = json.dumps(output_record)
                f.write(f"{output_record}\n")
        print(f"Wrote pickle to {output_file}.")


if __name__ == "__main__":
    app.run(main)



"""Migrations for source: https://github.com/mlcommons/croissant/issues/115."""

from typing import Any

from mlcroissant._src.core.issues import Issues
from mlcroissant._src.structure_graph.nodes.source import Source


def _is_in(elements: list[Any], name: str):
    for element in elements:
        if element.get("name") == name:
            return True
    return False


def _migrate_source(
    source: Source,
    record_sets: list[Any],
    file_sets: list[Any],
    file_objects: list[Any],
):
    if isinstance(source, dict):
        existing_transforms = source.get("applyTransform", [])
        if not isinstance(existing_transforms, list):
            existing_transforms = [existing_transforms]
        source["apply_transform"] = existing_transforms
    source = Source.from_jsonld(Issues(), source)
    reference = source.reference
    new_source = {}
    apply_transform = []
    for transform in list(source.apply_transform):
        if transform.format:
            apply_transform.append({"format": transform.format})
        elif transform.separator:
            apply_transform.append({"separator": transform.separator})
        elif transform.regex:
            apply_transform.append({"regex": transform.regex})
        elif transform.replace:
            apply_transform.append({"replace": transform.replace})
        else:
            raise ValueError("Impossible case")
    if apply_transform:
        if len(apply_transform) == 1:
            new_source["applyTransform"] = apply_transform[0]
        else:
            new_source["applyTransform"] = apply_transform
    if len(reference) == 2:
        origin, column = reference[0], reference[1]
    elif len(reference) == 1:
        origin, column = reference[0], None
    else:
        raise ValueError("Impossible case")
    # Check if source is a FileObject, FileSet or RecordSet
    if _is_in(record_sets, origin):
        if column is not None:
            new_source["field"] = f"{origin}/{column}"
    elif _is_in(file_sets, origin) or _is_in(file_objects, origin):
        new_source["distribution"] = origin
        if column == "filename" or column == "fullpath" or column == "content":
            new_source["ml:dataExtraction"] = {"fileProperty": column}
        elif column is not None:
            new_source["ml:dataExtraction"] = {"csvColumn": column}
    else:
        raise ValueError("Impossible case")
    return new_source


def up(json_ld):
    """Up migration that converts `source` to a structured source."""
    record_sets = json_ld.get("recordSet")
    if record_sets is None:
        return json_ld
    distribution = json_ld.get("distribution", [])
    if not isinstance(distribution, list):
        distribution = [distribution]
    file_sets = [el for el in distribution if "sc:FileSet" == el.get("@type", "")]
    file_objects = [el for el in distribution if "sc:FileObject" == el.get("@type", "")]
    for i, record_set in enumerate(record_sets):
        fields = record_set.get("field")
        if fields is None:
            continue
        for j, field in enumerate(fields):
            for source_field in ["source", "references"]:
                source = field.get(source_field)
                if source is not None:
                    new_source = _migrate_source(
                        source, record_sets, file_sets, file_objects
                    )
                    if new_source:
                        json_ld["recordSet"][i]["field"][j][source_field] = new_source
                sub_fields = field.get("subField")
                if sub_fields is None:
                    continue
                for k, sub_field in enumerate(sub_fields):
                    source = sub_field.get(source_field)
                    if source is not None:
                        new_source = _migrate_source(
                            source, record_sets, file_sets, file_objects
                        )
                        if new_source:
                            json_ld["recordSet"][i]["field"][j]["subField"][k][
                                source_field
                            ] = new_source
    return json_ld



"""Parse JSON operation."""

import jsonpath_rw
import pandas as pd

from mlcroissant._src.core.types import Json
from mlcroissant._src.structure_graph.nodes.field import Field


def parse_json_content(json: Json, fields: tuple[Field, ...]) -> pd.DataFrame:
    """Parsed all JSONs defined in the fields of RecordSet and outputs a pd.DF."""
    series = {}
    for field in fields:
        json_path = field.source.extract.json_path
        if json_path is None:
            continue
        jsonpath_expression = jsonpath_rw.parse(json_path)
        values = [match.value for match in jsonpath_expression.find(json)]
        series[json_path] = values
    return pd.DataFrame(series)



"""git module."""

import os

from absl import logging
from etils import epath

from mlcroissant._src.core.optional import deps
from mlcroissant._src.core.path import Path


def is_git_lfs_file(filepath: epath.Path) -> bool:
    """Returns whether a file a non-downloaded git-lfs file by checking its header.

    An optimization of this function would be to only read the file if there is a git
    repository in the distribution.
    """
    with filepath.open("rb") as file:
        # Only read the first line of the file. In the future, this could be a problem,
        # e.g. if we accept *.txt files and the file starts with the same header.
        first_line = file.readline()
        if first_line.startswith(b"version https://git-lfs.github.com/spec"):
            return True
    return False


def download_git_lfs_file(file: Path):
    """Downloads a specific git-lfs file within its repo."""
    # Path(filepath="/tmp/full/path.json", fullpath="path.json")
    # => working_dir = "/tmp/full"
    fullpath = os.fspath(file.fullpath)
    working_dir = os.fspath(file.filepath).rsplit(fullpath)[0]
    repo = deps.git.Git(working_dir)
    logging.info(
        "Downloading git-lfs file: %s in working dir: %s", fullpath, working_dir
    )
    try:
        repo.execute(["git", "lfs", "pull", "--include", fullpath])
    except deps.git.exc.GitCommandError as ex:
        raise RuntimeError(
            "Problem when launching `git lfs`. "
            "Possible problems: Have you installed git lfs "
            f"locally? Is '{fullpath}' a valid `git lfs` "
            "repository?"
        ) from ex



"""Extract operation module."""

import dataclasses
import logging
import os
import tarfile
import zipfile

from etils import epath

from mlcroissant._src.core.constants import EncodingFormat
from mlcroissant._src.core.constants import EXTRACT_PATH
from mlcroissant._src.core.path import get_fullpath
from mlcroissant._src.core.path import Path
from mlcroissant._src.operation_graph.base_operation import Operation
from mlcroissant._src.operation_graph.operations.download import get_hash
from mlcroissant._src.structure_graph.nodes.file_object import FileObject


def should_extract(encoding_format: str | None) -> bool:
    """Whether the encoding format should be extracted (zip or tar)."""
    return (
        encoding_format == EncodingFormat.TAR or encoding_format == EncodingFormat.ZIP
    )


def _extract_file(source: epath.Path, target: epath.Path) -> None:
    """Extracts the `source` file to `target`."""
    if zipfile.is_zipfile(source):
        with zipfile.ZipFile(source) as zip:
            zip.extractall(target)
    elif tarfile.is_tarfile(source):
        with tarfile.open(source) as tar:
            tar.extractall(target)
    else:
        raise ValueError(f"Unsupported compression method for file: {source}")


@dataclasses.dataclass(frozen=True, repr=False)
class Extract(Operation):
    """Extracts tar/zip and yields the extract Path."""

    node: FileObject

    def call(self, archive_file: Path) -> Path:
        """See class' docstring."""
        url = self.node.content_url
        assert url, "Content of URL for this node is None"
        hashed_url = get_hash(url)
        extract_dir = EXTRACT_PATH / hashed_url
        if not extract_dir.exists():
            _extract_file(archive_file.filepath, extract_dir)
        logging.info(
            "Found directory where data is extracted: %s", os.fspath(extract_dir)
        )
        return Path(
            filepath=extract_dir,
            fullpath=get_fullpath(extract_dir, EXTRACT_PATH),
        )



"""Global context."""

from __future__ import annotations

import dataclasses
import enum
from typing import Any, Mapping

from etils import epath
import networkx as nx

from mlcroissant._src.core.issues import Issues
from mlcroissant._src.core.rdf import Rdf


class CroissantVersion(enum.Enum):
    """Major and minor versions of the Croissant standard."""

    V_0_8 = "http://mlcommons.org/croissant/0.8"
    V_1_0 = "http://mlcommons.org/croissant/1.0"
    V_1_1 = "http://mlcommons.org/croissant/1.1"

    LATEST_VERSION = V_1_1

    @classmethod
    def from_jsonld(cls, ctx: Context, jsonld: Any) -> CroissantVersion:
        """Builds the class from the JSON-LD."""
        if isinstance(jsonld, CroissantVersion):
            return jsonld
        elif not jsonld:
            return CroissantVersion.V_0_8
        else:
            try:
                return CroissantVersion(jsonld)
            except ValueError:
                ctx.issues.add_error(
                    "conformsTo should be a string or a CroissantVersion. Got:"
                    f" {jsonld}"
                )
                return CroissantVersion.V_0_8

    def to_json(self) -> str | None:
        """Serializes back to JSON-LD."""
        if self == CroissantVersion.V_0_8:
            # In 0.8, the field conformsTo doesn't exist yet.
            return None
        return self.value

    def __lt__(self, other: CroissantVersion) -> bool:
        """Implements CroissantVersion.V_0_8 < CroissantVersion.V_1_0."""
        return self.value < other.value

    def __le__(self, other: CroissantVersion) -> bool:
        """Implements CroissantVersion.V_0_8 <= CroissantVersion.V_1_0."""
        return self.value <= other.value

    def __gt__(self, other: CroissantVersion) -> bool:
        """Implements CroissantVersion.V_1_0 > CroissantVersion.V_0_8."""
        return self.value > other.value

    def __ge__(self, other: CroissantVersion) -> bool:
        """Implements CroissantVersion.V_1_0 >= CroissantVersion.V_0_8."""
        return self.value >= other.value


@dataclasses.dataclass
class Context:
    """A global context in the application to store commonly used objects."""

    issues: Issues = dataclasses.field(default_factory=Issues)
    folder: epath.Path | None = None
    graph: nx.MultiDiGraph = dataclasses.field(
        default_factory=nx.MultiDiGraph, compare=False, init=False, hash=False
    )
    rdf: Rdf = dataclasses.field(default_factory=Rdf, hash=False)
    mapping: Mapping[str, epath.Path] = dataclasses.field(
        default_factory=dict, hash=False
    )
    conforms_to: CroissantVersion = CroissantVersion.V_1_0
    is_live_dataset: bool | None = None

    def __post_init__(self):
        """Standardizes conforms_to."""
        self.conforms_to = CroissantVersion.from_jsonld(self, self.conforms_to)

    def copy(self, **changes) -> Context:
        """Copies and replaces all changes."""
        return dataclasses.replace(self, **changes)

    def is_v0(self):
        """Whether the JSON-LD conforms to Croissant v0.8 or lower."""
        return self.conforms_to < CroissantVersion.V_1_0

    def is_v1_1(self):
        """Whether the JSON-LD conforms to Croissant v1.1 or higher."""
        return self.conforms_to >= CroissantVersion.V_1_1

    def is_latest_version(self):
        """Whether the JSON-LD conforms to the latest supported in mlcroissant."""
        return self.conforms_to == CroissantVersion.LATEST_VERSION



"""data_types module."""

import numpy as np
import pandas as pd
from rdflib import term

from mlcroissant._src.core import constants
from mlcroissant._src.core.constants import DataType
from mlcroissant._src.core.context import Context
from mlcroissant._src.core.issues import Issues
from mlcroissant._src.core.types import Json


def check_expected_type(issues: Issues, jsonld: Json, expected_type: str):
    """Checks that JSON-LD `jsonld` has "@type" == expected_type."""
    node_name = jsonld.get(constants.SCHEMA_ORG_NAME, "<unknown node>")
    node_type = jsonld.get("@type")
    if node_type != expected_type:
        issues.add_error(
            f'"{node_name}" should have an attribute "@type": "{expected_type}". Got'
            f" {node_type} instead."
        )


EXPECTED_DATA_TYPES: dict[term.URIRef, type] = {
    DataType.BOOL: bool,
    DataType.DATE: pd.Timestamp,
    DataType.FLOAT: float,
    DataType.FLOAT16: np.float16,
    DataType.FLOAT32: np.float32,
    DataType.FLOAT64: np.float64,
    DataType.INTEGER: int,
    DataType.INT8: np.int8,
    DataType.INT16: np.int16,
    DataType.INT32: np.int32,
    DataType.INT64: np.int64,
    DataType.TEXT: bytes,
    DataType.URL: bytes,
}


def data_types_from_jsonld(ctx: Context, data_types: Json):
    """Extracts DataType from its JSON-LD."""
    if isinstance(data_types, dict):
        data_type = data_types.get("@id")
        if isinstance(data_type, str):
            data_type = term.URIRef(data_type)
        return data_type
    elif isinstance(data_types, (str, term.URIRef)):
        return term.URIRef(data_types)
    elif isinstance(data_types, list):
        return [data_types_from_jsonld(ctx, d) for d in data_types]
    return []


def data_types_to_jsonld(ctx: Context, data_types: list[term.URIRef] | None):
    """Converts DataType to JSON-LD."""
    if data_types:
        return [ctx.rdf.shorten_value(data_type) for data_type in data_types]
    return None



"""Migration: Remove former RAI fields."""


def up(json_ld):
    """Up migration that removes former RAI fields."""
    if "@context" in json_ld:
        for field in ["dataCollection", "personalSensitiveInformation", "dataBiases"]:
            if field in json_ld["@context"]:
                del json_ld["@context"][field]
    return json_ld



"""Core utils to manipulate paths."""

import dataclasses
import os
import pathlib

from etils import epath


@dataclasses.dataclass
class Path:
    """Container class for paths in Croissant.

    Args:
        filepath: The full file path on disk.
        fullpath: `fullpath` in the sense of Croissant, i.e. the full path relative to
            the Croissant folders. This field may sound wrongly named, but it is named
            after the same JSON-LD field (see `fileProperty` in the Croissant specs).
            Example: for filepath=`${CROISSANT_CACHE}/downloads/my-dataset/my-file`, the
            fullpath is `my-dataset/my-file`.
    """

    filepath: epath.Path
    fullpath: pathlib.PurePath

    def __lt__(self, other):
        """Implements < for File allowing to sort a list[File]."""
        return os.fspath(self.filepath) < os.fspath(other.filepath)

    @property
    def filename(self) -> str:
        """The name of the file if it is a file."""
        return self.filepath.name


def get_fullpath(file: epath.Path, data_dir: epath.Path) -> pathlib.PurePath:
    """Returns the relative path relatively to `data_dir`."""
    # Path since the root of the dir.
    fullpath = os.fspath(file).replace(os.fspath(data_dir), "")
    # Remove the trailing slash.
    if fullpath.startswith("/"):
        fullpath = fullpath[1:]
    return pathlib.PurePath(fullpath)


def get_fullpaths(
    files: list[epath.Path], data_dir: epath.Path
) -> list[pathlib.PurePath]:
    """Returns the relative paths relatively to `data_dir`."""
    return [get_fullpath(file, data_dir) for file in files]



"""Field operation module."""

import dataclasses
import functools
import io
import logging
import os
import pathlib
import re
from typing import Any, Iterator

from etils import epath
import jsonpath_rw
import numpy as np
import pandas as pd
from rdflib import term

from mlcroissant._src.core.constants import DataType
from mlcroissant._src.core.context import Context
from mlcroissant._src.core.ml import bounding_box
from mlcroissant._src.core.optional import deps
from mlcroissant._src.operation_graph.base_operation import Operation
from mlcroissant._src.structure_graph.nodes.field import Field
from mlcroissant._src.structure_graph.nodes.record_set import RecordSet
from mlcroissant._src.structure_graph.nodes.source import FileProperty
from mlcroissant._src.structure_graph.nodes.source import Transform


@functools.cache
def _parse_jsonpath(json_path: str):
    """Memoizes jsonpath_rw.parse for better performances."""
    return jsonpath_rw.parse(json_path)


def _is_repeated_field(field: Field | None) -> bool | None:
    return isinstance(field, Field) and (field.repeated or field.is_array)


def _apply_transform_fn(value: Any, transform: Transform, field: Field) -> Any:
    """Applies one transform to `value`."""
    if _is_na(value):
        return value
    if transform.regex is not None:
        source_regex = re.compile(transform.regex)
        if isinstance(value, pathlib.PurePath):
            value = os.fspath(value)
        match = source_regex.match(value)
        if match is None:
            logging.warning(f"Could not match {source_regex} in {value}")
            return value
        for group in match.groups():
            if group is not None:
                return group
    elif transform.json_path is not None:
        jsonpath_expression = _parse_jsonpath(transform.json_path)
        if matches := jsonpath_expression.find(value):
            return next(match.value for match in matches)
        return None
    elif transform.format is not None:
        if field.data_type is pd.Timestamp:
            return pd.Timestamp(value).strftime(transform.format)
        else:
            raise ValueError(f"`format` only applies to dates. Got {field.data_type}")
    elif transform.separator is not None:
        return value.split(transform.separator)
    return value


def apply_transforms_fn(value: Any, field: Field, repeated: bool = False) -> Any:
    """Applies all transforms in `source` to `value`."""
    source = field.source
    if source is None:
        return value
    transforms = source.transforms
    for transform in transforms:
        if repeated and isinstance(value, (list, np.ndarray)):
            value = [_apply_transform_fn(v, transform, field) for v in value]
        else:
            value = _apply_transform_fn(value, transform, field)
    return value


def _is_na(value: Any) -> bool:
    return not isinstance(value, (list, np.ndarray)) and pd.isna(value)


def _cast_value(ctx: Context, value: Any, data_type: type | term.URIRef | None):
    """Casts the value `value` to the desired target data type `data_type`."""
    if _is_na(value):
        return value
    elif data_type == DataType.IMAGE_OBJECT:
        if isinstance(value, deps.PIL_Image.Image):
            return value
        elif isinstance(value, bytes):
            return deps.PIL_Image.open(io.BytesIO(value))
        else:
            raise ValueError(f"Type {type(value)} is not accepted for an image.")
    elif data_type == DataType.AUDIO_OBJECT:
        output = deps.librosa.load(io.BytesIO(value))
        return output
    elif data_type == DataType.BOUNDING_BOX:  # pytype: disable=wrong-arg-types
        return bounding_box.parse(value)
    elif not isinstance(data_type, type):
        raise ValueError(f"No special case for type {data_type}.")
    elif isinstance(value, list) or isinstance(value, np.ndarray):
        return [_cast_value(ctx=ctx, value=v, data_type=data_type) for v in value]
    elif data_type == bytes and not isinstance(value, bytes):
        return _to_bytes(value)
    else:
        return data_type(value)


def _to_bytes(value: Any) -> bytes:
    """Casts the value `value` to bytes."""
    if isinstance(value, bytes):
        return value
    elif isinstance(value, str):
        return value.encode("utf-8")
    elif isinstance(value, int) or isinstance(value, float):
        return str(value).encode("utf-8")
    else:
        return bytes(value)


def _read_file(row: pd.Series) -> pd.Series:
    """Reads a binary file."""
    path = row[FileProperty.filepath]
    content = epath.Path(path).open("rb").read()
    return pd.Series({**row, FileProperty.content: content})


def _extract_lines(row: pd.Series) -> pd.Series:
    """Reads a file line-by-line and outputs a named pd.Series of the lines."""
    path = epath.Path(row[FileProperty.filepath])
    lines = path.open("rb").read().splitlines()
    return pd.Series({
        **row, FileProperty.lines: lines, FileProperty.lineNumbers: range(len(lines))
    })


def _extract_value(df: pd.DataFrame, field: Field) -> pd.DataFrame:
    """Extracts the value according to the field rules."""
    source = field.source
    column_name = source.get_column()
    if column_name in df:
        return df
    elif source.extract.file_property == FileProperty.content:
        return df.apply(_read_file, axis=1)
    elif source.extract.file_property in [FileProperty.lines, FileProperty.lineNumbers]:
        df = df.apply(_extract_lines, axis=1)
        return df.explode(
            column=[FileProperty.lines, FileProperty.lineNumbers], ignore_index=True
        )
    return df


def _populate_repeated_nested_subfield(
    value: Any, field: Field, result: dict[str, Any]
) -> dict[str, Any]:
    """Populates result with a field's nested subfields."""
    if not field.parent:
        raise ValueError(
            "Nested subfields can only be populated when the parent field exists!"
        )
    parent_id = field.parent.id
    existing_values = result.get(parent_id, None)
    if existing_values:
        if not _is_na(value) and len(value) != len(existing_values):
            raise ValueError(
                f"Lenghts of {field.id} doesn't match  already stored items for "
                f" {parent_id}"
            )
        for i in range(len(existing_values)):
            existing_values[i][field.id] = None if _is_na(value) else value[i]
    else:
        result[parent_id] = (
            [{field.id: None}] if _is_na(value) else [{field.id: v} for v in value]
        )
    return result


@dataclasses.dataclass(frozen=True, repr=False)
class ReadFields(Operation):
    """Reads fields in a RecordSet from a Pandas DataFrame and applies transformations.

    ReadFields.__call__() yields dict-shaped records with the proper transformations,
    types and name.
    """

    node: RecordSet

    def _get_fields(self, fields: list[Field]):
        """Extracts all leaves fields (i.e., including subFields)."""
        all_fields: list[Field] = []
        for field in fields:
            if field.sub_fields:
                all_fields.extend(self._get_fields(field.sub_fields))
            else:
                all_fields.append(field)
        return all_fields

    def call(self, df: pd.DataFrame) -> Iterator[dict[str, Any]]:
        """See class' docstring."""
        if self.node.data:
            # The RecordSet has `data`, so we directly yield from the dataframe.
            for _, row in df.iterrows():
                yield dict(row)
            return
        fields = self._get_fields(self.node.fields)
        for field in fields:
            df = _extract_value(df, field)

        def _get_result(row):
            """Returns a record parsed as a dictionary of fields."""
            result: dict[str, Any] = {}
            for field in fields:
                source = field.source
                column = source.get_column()
                assert column in df, (
                    f'Column "{column}" does not exist. Inspect the ancestors of the'
                    f" field {field} to understand why. Possible fields: {df.columns}"
                )
                is_repeated = field.repeated or _is_repeated_field(field.parent)
                value = apply_transforms_fn(
                    row[column], field=field, repeated=is_repeated
                )
                if _is_na(value):
                    value = None
                elif is_repeated:
                    value = [
                        _cast_value(self.node.ctx, v, field.data_type) for v in value
                    ]
                else:
                    value = _cast_value(self.node.ctx, value, field.data_type)

                if self.node.ctx.is_v0():
                    result[field.name] = value
                else:
                    if field in self.node.fields:
                        result[field.id] = value
                    else:
                        # Repeated nested sub-fields render as a list of dictionaries.
                        if _is_repeated_field(field.parent):
                            result = _populate_repeated_nested_subfield(
                                value=value, field=field, result=result
                            )
                        # Non-repeated subfields render as a single dictionary.
                        else:
                            parent_id = field.parent.id
                            if parent_id not in result:
                                result[parent_id] = {}
                            result[parent_id][field.id] = value
            return result

        chunk_size = 100
        for i in range(0, len(df), chunk_size):
            yield from df[i : i + chunk_size].apply(_get_result, axis=1)



"""Migration: `ml:` -> `cr:`."""

import json


def up(json_ld):
    """Up migration for `ml:` -> `cr:`."""
    json_ld = json.dumps(json_ld)
    json_ld = json_ld.replace("ml:", "cr:")
    json_ld = json.loads(json_ld)
    json_ld["@context"]["cr"] = "http://mlcommons.org/croissant/"
    if "ml" in json_ld["@context"]:
        del json_ld["@context"]["ml"]
    return json_ld



"""Module to execute operations."""

from __future__ import annotations

from collections.abc import Mapping
import concurrent.futures
import functools
import json
import sys
from typing import Any, Generator

from absl import logging
import networkx as nx

from mlcroissant._src.core.issues import GenerationError
from mlcroissant._src.operation_graph.base_operation import Operation
from mlcroissant._src.operation_graph.base_operation import Operations
from mlcroissant._src.operation_graph.operations import FilterFiles
from mlcroissant._src.operation_graph.operations import ReadFields
from mlcroissant._src.operation_graph.operations.download import Download
from mlcroissant._src.operation_graph.operations.read import Read

ElementWithIndex = tuple[int, Any]


def execute_downloads(operations: Operations):
    """Executes all the downloads in the graph of operations."""
    downloads = [
        operation for operation in operations.nodes if isinstance(operation, Download)
    ]
    with concurrent.futures.ThreadPoolExecutor() as executor:
        for download in downloads:
            executor.submit(download)


def _order_relevant_operations(
    operations: Operations, record_set_id: str
) -> list[Operation]:
    """Orders all relevant operations for the RecordSet."""
    # ReadFields linked to the `record_set_name`.
    read_fields_operation = next(
        (
            operation
            for operation in operations.nodes
            if isinstance(operation, ReadFields)
            and operation.node.uuid == record_set_id
        )
    )
    ancestors = set(nx.ancestors(operations, read_fields_operation))
    # Return ReadFields and all its ancestors
    return [
        operation
        for operation in nx.topological_sort(operations)
        # If the operation is not a needed operation to compute `record_set`, skip it:
        if operation in ancestors or operation == read_fields_operation
    ]


def execute_operations_sequentially(record_set: str, operations: Operations):
    """Executes operation and yields results according to the graph of operations."""
    for operation in _order_relevant_operations(operations, record_set):
        try:
            logging.info("Executing %s", operation)
            if isinstance(operation, ReadFields) and record_set == operation.node.uuid:
                # This is the target RecordSet, so we can yield records. We don't keep
                # the output in memory, so that we can yield from bigger-than-memory
                # generators.
                yield from operation(set_output_in_memory=False)
            else:
                operation(set_output_in_memory=True)
        except Exception as e:
            raise GenerationError(
                "An error occured during the sequential generation of the dataset, more"
                f" specifically during the operation {operation}"
            ) from e


def execute_operations_in_streaming(
    record_set: str,
    operations: Operations,
    list_of_operations: list[Operation] | None = None,
    result: Any = None,
):
    """Executes operation and streams results when reading files.

    This allows to stream from operations that return a list (e.g., FilterFiles) in
    order not to block on long operations. Instead of downloading the entire dataset,
    we only download the needed files, yield element, then proceed to the next file.
    """
    if list_of_operations is None:
        list_of_operations = _order_relevant_operations(operations, record_set)
    for i, operation in enumerate(list_of_operations):
        try:
            if isinstance(operation, ReadFields):
                if operation.node.uuid != record_set:
                    continue
                yield from operation.call(result)
                return
            elif isinstance(operation, Read):
                # At this stage `result` can be either a Path or a list of Paths.
                if not isinstance(result, list):
                    result = [result]

                def read_all_files():
                    for file in result:
                        # Read files separately and keep executing subsequent operations
                        logging.info("Executing %s", operation)
                        yield from execute_operations_in_streaming(
                            record_set=record_set,
                            operations=operations,
                            list_of_operations=list_of_operations[i + 1 :],
                            result=operation.call(file),
                        )

                yield from read_all_files()
                return
            else:
                logging.info("Executing %s", operation)
                result = operation.call(result)
        except Exception as e:
            raise GenerationError(
                "An error occured during the streaming generation of the dataset, more"
                f" specifically during the operation {operation}"
            ) from e


def execute_operations_in_beam(
    record_set: str,
    operations: Operations,
    filters: Mapping[str, Any] | None = None,
):
    """See ReadFromCroissant docstring."""
    import apache_beam as beam

    list_of_operations = _order_relevant_operations(operations, record_set)
    target = list_of_operations[-1]
    if not isinstance(target, ReadFields):
        raise ValueError("the target MUST correspond to the RecordSet to generate.")

    # We use the FilterFiles operation to parallelize operations. If there's no
    # FilterFile operation, we set it to `target`.
    filter_files = _find_filter_files(operations, target)
    stage_prefix = f"{record_set} " + json.dumps(filters) if filters else "no filter"

    # In memory = all operations that are not between FilterFiles and the target.
    # In Beam = all operations that are between FilterFiles and the target.
    # Example for a Hugging Face dataset:
    #
    #                         Download(repo)                   Data(splits)
    #                              │                                 │
    #                              ▼                                 │
    #                   FilterFiles(parquet-files)                   │
    #                              │                                 │
    #                              ▼                                 │
    #                 ┌────────────────────────────┐                 ▼
    #                 │      Read(parquet-files)   │         ReadFields(splits)
    #           Those │            │               │                 |
    #      operations │            ▼               │                 |
    #             are │        Join(default)   ◄─────────────────────┘
    #        executed │            │               │
    #     in parallel │            ▼               │
    #         in Beam │    ReadFields(default)     │
    #                 └────────────────────────────┘

    operations_in_memory: list[Operation] = []
    operations_in_beam: list[Operation] = []
    for paths in nx.all_simple_paths(operations, source=filter_files, target=target):
        for operation in paths:
            if operation != filter_files:
                operations_in_beam.append(operation)
    for operation in list_of_operations:
        if operation not in operations_in_beam:
            operations_in_memory.append(operation)

    # Execute all in-memory operations.
    for operation in operations_in_memory:
        # If there is no FilterFiles, we return the PCollection without parallelization.
        if operation == target:
            return beam.Create([(0, *operation.inputs)]) | _beam_operation_with_index(
                operation, sys.maxsize, stage_prefix
            )
        else:
            operation(set_output_in_memory=True)

    files = filter_files.output  # even for large datasets, this can be handled in RAM.
    # We first shard by file and assign a shard_index.
    pipeline = beam.Create(enumerate(files))
    num_shards = len(files)
    if not num_shards:
        raise ValueError(
            f"Empty {record_set=}. No files found for filters={json.dumps(filters)}"
        )

    # We don't know in advance the number of records per shards. So we just allocate the
    # maximum number which is `sys.maxsize // num_shards`. Taking the practical case of
    # large evenly distributed datasets on HuggingFace, we can compute the following:

    # num_shards = number of Parquet files per config on Hugging Face < 10 billion files
    # max_shard_size ~ 1 billion records per Parquet files

    # So it seems we can run with this trick without too many problems. We still trigger
    # a ValueError below if the error arises, and we ask the user to open a bug. A real
    # solution to this problem would be to compute the shard_sizes in parallel of
    # generating the records.
    # TODO(https://github.com/mlcommons/croissant/issues/732): Compute shard_sizes
    # explicitly instead of relying on max_shard_size.
    max_shard_size = sys.maxsize // num_shards
    for operation in operations_in_beam:
        beam_operation = _beam_operation_with_index(
            operation, max_shard_size, stage_prefix
        )
        pipeline |= beam_operation
    return pipeline


def _add_global_index(
    element_with_index: ElementWithIndex,
    operation: Operation,
    max_shard_size: int,
) -> Generator[ElementWithIndex, None, None]:
    """Computes the global index given the shard size."""
    shard_index, element = element_with_index
    for index_in_shard, result in enumerate(operation.call(element)):
        if index_in_shard >= max_shard_size:
            raise ValueError(
                "WARNING: This was very unlikely, but it seems we just hit this limit"
                " in the code. Find another way to optimize execute_operations_in_beam."
                " Please, open a PR on GitHub to make the maintainers aware of this"
                " issue. A fix is to compute the actual shard_sizes rather than relying"
                " on a heuristic (see comments above in code)."
            )
        new_index = max_shard_size * shard_index + index_in_shard
        yield (new_index, result)


def _pass_index(index: int, element: Any, operation: Operation) -> ElementWithIndex:
    """Passes the index to the next operation while executing the operation."""
    return (index, operation.call(element))


def _beam_operation_with_index(
    operation: Operation, max_shard_size: int, stage_prefix: str | None
):
    import apache_beam as beam

    if isinstance(operation, ReadFields):
        return (
            f"{stage_prefix} {operation.node.uuid}: compute the global index."
            >> beam.ParDo(
                functools.partial(
                    _add_global_index,
                    operation=operation,
                    max_shard_size=max_shard_size,
                )
            )
        )
    else:
        return (
            f"{stage_prefix} {operation.node.uuid}: pass the index to the next"
            " operation."
            >> beam.MapTuple(functools.partial(_pass_index, operation=operation))
        )


def _find_filter_files(operations: Operations, target: ReadFields) -> Operation:
    """Finds a FilterFiles operation to parallelize on."""
    filter_files_operations = [
        operation for operation in operations if isinstance(operation, FilterFiles)
    ]
    match len(filter_files_operations):
        case 0:
            return target
        case 1:
            filter_files = filter_files_operations[0]
            if not nx.has_path(operations, source=filter_files, target=target):
                raise NotImplementedError("No path between FilterFiles and the target")
            return filter_files
        case _:
            # If joined branches also have FilterFiles operations, this case is not
            # handled yet. If you have this issue, please open an issue on GitHub.
            raise NotImplementedError("Found several FilterFiles - not handled yet.")



"""Download operation module."""

import base64
import dataclasses
import hashlib
import logging
import os
import time
import urllib.parse

from etils import epath
import requests
import tqdm

from mlcroissant._src.core import constants
from mlcroissant._src.core.constants import EncodingFormat
from mlcroissant._src.core.optional import deps
from mlcroissant._src.core.path import get_fullpath
from mlcroissant._src.core.path import Path
from mlcroissant._src.core.url import is_url
from mlcroissant._src.operation_graph.base_operation import Operation
from mlcroissant._src.structure_graph.nodes.file_object import FileObject

_CHUNK_SIZE = 1024
_GITHUB_GIT = "https://github.com"
_GITLAB_GIT = "https://gitlab.com"
_HUGGING_FACE_GIT = "https://huggingface.co/datasets"
_SUPPORTED_HOSTS = [
    _GITHUB_GIT,
    _GITLAB_GIT,
    _HUGGING_FACE_GIT,
]


def get_hash(url: str) -> str:
    """Retrieves the sha256 hash of an URL."""
    return hashlib.sha256(url.encode()).hexdigest()


def get_download_filepath(node: FileObject) -> epath.Path:
    """Retrieves the download filepath of an URL."""
    ctx = node.ctx
    if node.id in ctx.mapping:
        return ctx.mapping[node.id]
    url = node.content_url
    if url and not is_url(url) and not node.contained_in:
        if ctx.folder is None:
            raise ValueError(f"Could not find node folder={ctx.folder}")
        filepath = ctx.folder / url
        assert filepath.exists(), (
            f'In node "{node.uuid}", file "{url}" is either an invalid URL'
            " or an invalid path."
        )
        # No need to download local files
        return filepath
    assert url, "Cannot get hashed URL for None."
    hashed_url = get_hash(url)
    constants.DOWNLOAD_PATH.mkdir(parents=True, exist_ok=True)
    return constants.DOWNLOAD_PATH / f"croissant-{hashed_url}"


def insert_credentials(url: str, username: str | None, password: str | None) -> str:
    """Inserts credentials in the URL for an HTTP authentication.

    Args:
        url: The Git URL.
        username: The username to use for the HTTPS authentication.
        password: The password to use for the HTTPS authentication.

    Returns:
        The URL populated with the given credentials. For example:
        `https://{username}:{password}@github.com/account/repo`.
    """
    https = "https://"
    if not url.startswith(https):
        raise ValueError("use secured HTTPS protocol for {url}.")
    if username is None and password is None:
        return url
    if username is None or password is None:
        raise ValueError(
            f"please provide both {constants.CROISSANT_GIT_USERNAME} and"
            f" {constants.CROISSANT_GIT_PASSWORD}."
        )
    username = urllib.parse.quote(username)
    password = urllib.parse.quote(password)
    https_with_credentials = f"{https}{username}:{password}@"
    return url.replace(https, https_with_credentials)


def extract_git_info(full_url: str) -> tuple[str, str | None]:
    """Extracts git-related information.

    Args:
        full_url: The full git URL for an HTTP clone.

    Returns:
        A tuple containing the base URL of the repository to clone and the refs to
        checkout if any is provided.
    """
    full_url = os.fspath(full_url)
    if full_url.startswith(_GITHUB_GIT) or full_url.startswith(_GITLAB_GIT):
        return full_url, None
    if full_url.startswith(_HUGGING_FACE_GIT):
        splits = full_url.rsplit("/tree/refs%2F", 1)
        if len(splits) == 1:
            # There is no refs in the URL: returning None.
            return splits[0], None
        else:
            url, refs = splits
            return url, f"refs/{urllib.parse.unquote(refs)}"
    else:
        raise ValueError(
            f"unknown git host: {full_url}. Supported git hosts:"
            f" {_SUPPORTED_HOSTS}. Contact the Croissant team to support more hosts."
        )


def _get_hash_algorithm(file_object: FileObject):
    if file_object.md5:
        return hashlib.md5()
    elif file_object.sha256:
        return hashlib.sha256()
    else:
        raise ValueError("md5 and sha256 are not specified.")


def get_basic_auth_from_env() -> tuple[str, str] | None:
    """Determines a Basic Auth tuple from the environment variables.

    This method determines the username and password for the auth tuple from the
    `CROISSANT_BASIC_AUTH_USERNAME` and `CROISSANT_BASIC_AUTH_PASSWORD` env variables.

    Returns:
        The Basic Auth tuple if the env variables are configured properly. Otherwise, it
        returns None.
    """
    username = os.environ.get(constants.CROISSANT_BASIC_AUTH_USERNAME)
    password = os.environ.get(constants.CROISSANT_BASIC_AUTH_PASSWORD)
    return None if username is None or password is None else (username, password)


@dataclasses.dataclass(frozen=True, repr=False)
class Download(Operation):
    """Downloads from a URL to the disk."""

    node: FileObject

    def _check_hash(self, filepath: epath.Path):
        if filepath.is_dir():
            return
        ctx = self.node.ctx
        # For live datasets which do not specify checksums, we do not check the hash.
        if ctx.is_live_dataset and not (self.node.md5 or self.node.sha256):
            logging.info(
                "Checksums not provided for a live dataset, no hash will be checked."
            )
            return
        hash = _get_hash_algorithm(self.node)
        with filepath.open("rb") as f:
            while chunk := f.read(_CHUNK_SIZE):
                hash.update(chunk)
        # The hash from the Croissant JSON-LD can be in either hex or base64,
        # so let's check both
        expected_hash = getattr(self.node, hash.name)
        # First, try hex as that's likely more common
        hex_hash = hash.hexdigest()
        if hex_hash == expected_hash:
            return
        # Next, try base64 as a fallback
        base64_hash = base64.b64encode(bytes.fromhex(hex_hash)).decode()
        if base64_hash == expected_hash:
            return

        logging.info(
            "Hash of downloaded file is not identical with reference in metadata.json"
        )
        # In v0.8 only, hashes were not checked.
        if not ctx.is_v0():
            raise ValueError(
                f"Hash of downloaded file {filepath} is not identical with the"
                f" reference in the Croissant JSON-LD. Expected: {expected_hash} -"
                f" Got: {hex_hash} (hex) / {base64_hash} (base64)"
            )

    def _download_from_http(self, filepath: epath.Path):
        content_url = self.node.content_url
        assert content_url, "Content URL is None."
        response = requests.get(
            content_url, stream=True, timeout=10, auth=get_basic_auth_from_env()
        )
        response.raise_for_status()
        total = int(response.headers.get("Content-Length", 0))
        with filepath.open("wb") as file, tqdm.tqdm(
            desc=f"Downloading {content_url}...",
            total=total,
            unit="iB",
            unit_scale=True,
            unit_divisor=1024,
        ) as bar:
            for data in response.iter_content(chunk_size=_CHUNK_SIZE):
                size = file.write(data)
                bar.update(size)

    def _download_from_git(self, filepath: epath.Path):
        username = os.environ.get(constants.CROISSANT_GIT_USERNAME)
        password = os.environ.get(constants.CROISSANT_GIT_PASSWORD)
        # GIT_LFS_SKIP_SMUDGE allows to not download git-lfs files by default. Those
        # files may be big, so we want to download them just-in-time (see `Read`):
        os.environ["GIT_LFS_SKIP_SMUDGE"] = "1"

        assert self.node.content_url, "Cannot extract git-related information for None."
        url, refs = extract_git_info(self.node.content_url)
        url = insert_credentials(url, username, password)
        repo = deps.git.Repo.clone_from(url, filepath)
        # Hugging Face uses https://git-scm.com/book/en/v2/Git-Internals-Git-References.
        if refs:
            branch_name = f"branch_{time.time()}"  # Branch name with no conflict
            repo.remote().fetch(f"{refs}:{branch_name}")
            repo.branches[branch_name].checkout()

    def call(self, *args) -> Path:
        """See class' docstring."""
        del args  # unused
        filepath = get_download_filepath(self.node)
        if not filepath.exists():
            if self.node.encoding_format == EncodingFormat.GIT:
                self._download_from_git(filepath)
            else:
                self._download_from_http(filepath)
        self._check_hash(filepath)
        logging.info(
            "File %s is downloaded to %s", self.node.content_url, os.fspath(filepath)
        )
        return Path(
            filepath=filepath,
            fullpath=get_fullpath(filepath, constants.DOWNLOAD_PATH),
        )



"""A minimal PyTorch adapter on top of mlcroissant.

The DataPipe abstraction is currently used to enable composing more complex
datasets.
"""

import enum
from typing import Any, Dict, Optional

from mlcroissant._src.core.optional import deps
from mlcroissant._src.datasets import Dataset

try:
    dp = deps.torchdata_datapipes
except ModuleNotFoundError:
    dp = None

INSTALL_MESSAGE = "torchdata is not installed and is a dependency."


class LoaderSpecificationDataType(enum.Enum):
    """Enum representing data type specification."""

    INFER = "infer"
    UTF8 = "utf8"


# Map from name to DataType
LoaderSpecificationTypes = Dict[str, Optional[LoaderSpecificationDataType]]


def infer_data_type(val: Any) -> Optional[LoaderSpecificationDataType]:
    """Automatically infer LoaderSpecificationDataType from val."""
    if isinstance(val, bytes):
        return LoaderSpecificationDataType.UTF8

    return None


def apply_data_type_transformation(
    val: Any, data_type: Optional[LoaderSpecificationDataType]
) -> Any:
    """Transform val according to data_type."""
    if data_type == LoaderSpecificationDataType.INFER:
        # Infer and Forward
        data_type = infer_data_type(val)
        return apply_data_type_transformation(val, data_type)

    if data_type == LoaderSpecificationDataType.UTF8:
        return val.decode("utf-8")

    return val


class LoaderFactory:
    """Used to create loaders and get metadata."""

    def __init__(self, jsonld: str):
        """Initialize LoaderFactory with a Croissant file."""
        self.jsonld = jsonld

    def _get_row_processor(self, specification: LoaderSpecificationTypes):
        """Remap columns types to desired type."""

        def row_processor(row):
            for k, v in specification.items():
                row[k] = apply_data_type_transformation(row[k], v)
            return row

        return row_processor

    def as_datapipe(
        self,
        record_set: str,
        specification: Optional[LoaderSpecificationTypes] = None,
    ) -> Any:
        """Load the record set as a DataPipe."""
        if dp is None:
            raise NotImplementedError(INSTALL_MESSAGE)

        dataset = Dataset(jsonld=self.jsonld)
        records = dataset.records(record_set=record_set)
        datapipe = dp.iter.IterableWrapper(records)
        if specification:
            row_processor = self._get_row_processor(specification)
            datapipe = datapipe.map(row_processor)

        # TODO: Mark the return type as dp.iter.IterDataPipe
        return datapipe



"""Tests for Metadata."""

import copy
import datetime
from typing import Any
from unittest import mock

import pytest
from rdflib.namespace import SDO

from mlcroissant._src.core import constants
from mlcroissant._src.core.context import Context
from mlcroissant._src.core.context import CroissantVersion
from mlcroissant._src.core.issues import ValidationError
from mlcroissant._src.structure_graph.base_node import Node
from mlcroissant._src.structure_graph.nodes.creative_work import CreativeWork
from mlcroissant._src.structure_graph.nodes.field import Field
from mlcroissant._src.structure_graph.nodes.metadata import Metadata
from mlcroissant._src.structure_graph.nodes.record_set import RecordSet
from mlcroissant._src.tests.nodes import create_test_node
from mlcroissant._src.tests.versions import parametrize_conforms_to


def test_checks_are_performed():
    with mock.patch.object(
        Node, "assert_has_mandatory_properties"
    ) as mandatory_mock, mock.patch.object(
        Node, "assert_has_optional_properties"
    ) as optional_mock, mock.patch.object(
        Node, "validate_name"
    ) as validate_name_mock:
        ctx = Context(conforms_to=CroissantVersion.V_0_8)
        create_test_node(Metadata, name="field_name", id="field_id", ctx=ctx)
        mandatory_mock.assert_called_once_with("name")
        optional_mock.assert_called_once_with(
            "cite_as", "date_published", "license", "version"
        )
        validate_name_mock.assert_called_once()


@parametrize_conforms_to()
@pytest.mark.parametrize("version", [1, 1.0, "1.0.0"])
def test_from_jsonld(conforms_to: CroissantVersion, version: Any):
    ctx = Context(conforms_to=conforms_to)
    jsonld = {
        "@type": constants.SCHEMA_ORG_DATASET,
        constants.SCHEMA_ORG_NAME: "foo",
        constants.SCHEMA_ORG_DESCRIPTION: "bar",
        constants.ML_COMMONS_CITE_AS(ctx): "bixtex: citation",
        constants.DCTERMS_CONFORMS_TO: conforms_to.value,
        # Dates can be datetimes...
        constants.SCHEMA_ORG_DATE_CREATED: "1990-02-01",
        # ...or dates can be string...
        constants.SCHEMA_ORG_DATE_MODIFIED: "1990-02-02",
        # ...or dates can be datetime.dates.
        constants.SCHEMA_ORG_DATE_PUBLISHED: "1990-02-03",
        constants.SCHEMA_ORG_LICENSE: [
            {
                "@type": SDO.CreativeWork,
                SDO.name: "U.S. Government Works",
                SDO.url: "https://www.usa.gov/government-works/",
            },
            "License",
        ],
        constants.SCHEMA_ORG_URL: "https://mlcommons.org",
        constants.SCHEMA_ORG_VERSION: version,
        constants.ML_COMMONS_IS_LIVE_DATASET(ctx): False,
    }
    metadata = Metadata.from_jsonld(ctx, jsonld)
    assert metadata.name == "foo"
    assert metadata.description == "bar"
    assert metadata.date_created == datetime.datetime(1990, 2, 1, 0, 0)
    assert metadata.date_modified == datetime.datetime(1990, 2, 2, 0, 0)
    assert metadata.date_published == datetime.datetime(1990, 2, 3, 0, 0)
    assert len(metadata.license) == 2
    assert isinstance(metadata.license[0], CreativeWork)
    assert metadata.license[0].name == "U.S. Government Works"
    assert metadata.license[0].url == "https://www.usa.gov/government-works/"
    assert metadata.license[1] == "License"
    assert metadata.ctx.is_live_dataset == False
    assert metadata.url == "https://mlcommons.org"
    assert metadata.version == "1.0.0"
    assert not ctx.issues.errors
    assert not ctx.issues.warnings


@pytest.mark.parametrize(
    ["version", "expected_error"],
    [
        [
            ["1.2.3"],
            "The version should be a string or a number. Got: <class 'list'>.",
        ],
    ],
)
def test_invalid_version(version, expected_error):
    ctx = Context()
    with pytest.raises(ValidationError, match=rf"{expected_error}"):
        Metadata(ctx, name="foo", version=version)


@pytest.mark.parametrize(
    ["version", "expected_warning"],
    [
        [
            "1.2.x",
            "Version doesn't follow MAJOR.MINOR.PATCH: 1.2.x.",
        ],
        [
            "...123",
            "Version doesn't follow MAJOR.MINOR.PATCH: ...123.",
        ],
        [
            "a.b.c",
            "Version doesn't follow MAJOR.MINOR.PATCH: a.b.c",
        ],
    ],
)
def test_warning_version(version, expected_warning):
    ctx = Context()
    metadata = Metadata(ctx, name="foo", version=version)
    assert any(expected_warning in warning for warning in metadata.ctx.issues.warnings)


@pytest.mark.parametrize(
    ["version", "expected_version"],
    [
        ["1", "1"],
        ["1.2", "1.2"],
        ["1.2.3", "1.2.3"],
        ["1.2.3-foo+bar", "1.2.3-foo+bar"],
        [1, "1.0.0"],
        [1.2, "1.2.0"],
        ["thisisanarbitraryversion", "thisisanarbitraryversion"],
    ],
)
def test_valid_version(version, expected_version):
    ctx = Context()
    metadata = Metadata(ctx, name="foo", version=version)
    assert metadata.version == expected_version


def test_issues_in_metadata_are_shared_with_children():
    with pytest.raises(ValidationError, match="is mandatory, but does not exist"):
        ctx = Context(conforms_to=CroissantVersion.V_0_8)
        Metadata(
            name="name",
            description="description",
            url="https://mlcommons.org",
            version="1.0.0",
            # We did not specify the RecordSet's name. Hence the exception above:
            record_sets=[
                RecordSet(id="record-set", description="description", ctx=ctx)
            ],
        )


def test_metadata_can_be_deep_copied():
    metadata = Metadata(name="foo")
    # PyTorch DataPipes requries copy.deepcopy:
    copied_metadata = copy.deepcopy(metadata)
    assert copied_metadata.name == metadata.name == "foo"
    assert copied_metadata is not metadata


def test_validate_license():
    # Disabling PyType in case someone doesn't use PyType.
    assert Metadata(name="foo").license is None
    assert Metadata(name="foo", license=None).license is None
    assert Metadata(name="foo", license="mit").license == [
        "mit"
    ]  # pytype: disable=wrong-arg-types
    assert Metadata(name="foo", license=["apache-2.0", "mit"]).license == [
        "apache-2.0",
        "mit",
    ]
    with pytest.raises(ValidationError, match="License should be a list of str"):
        Metadata(name="foo", license=42)  # pytype: disable=wrong-arg-types


def test_predecessors_are_propagated():
    field = Field(
        id="records/name",
        name="records/name",
        data_types=constants.DataType.TEXT,
    )
    record_set = RecordSet(
        id="records",
        fields=[field],
        data=[{"records/name": "train"}, {"records/name": "test"}],
    )
    Metadata(name="dummy", record_sets=[record_set])
    assert field.predecessors == {record_set}


def test_parents_are_defined():
    subfield = Field(
        id="records/name/subfield",
        name="records/name/subfield",
        data_types=constants.DataType.TEXT,
    )
    field = Field(
        id="records/name",
        name="records/name",
        sub_fields=[subfield],
    )
    record_set = RecordSet(
        id="records",
        fields=[field],
        data=[{"records/name": "train"}, {"records/name": "test"}],
    )
    Metadata(name="dummy", record_sets=[record_set])
    assert {record_set.uuid, field.uuid}.issubset(
        set([p.uuid for p in subfield.parents])
    )
    assert {record_set.uuid}.issubset(set([p.uuid for p in field.parents]))



"""Migrations for containedIn: https://github.com/mlcommons/croissant/issues/115."""


def _migrate_contained_in(contained_in: str | list[str]):
    if isinstance(contained_in, list):
        return [_migrate_contained_in(el) for el in contained_in]
    elif isinstance(contained_in, str):
        assert contained_in.startswith("#{")
        assert contained_in.endswith("}")
        return contained_in.replace("#{", "").replace("}", "")
    else:
        raise ValueError("Impossible case")


def up(json_ld):
    """Up migration that converts `containedIn` to the new format."""
    distributions = json_ld.get("distribution", [])
    if not isinstance(distributions, list):
        distributions = [distributions]
    for i, distribution in enumerate(distributions):
        if "containedIn" not in distribution:
            continue
        new_contained_in = _migrate_contained_in(distribution["containedIn"])
        json_ld["distribution"][i]["containedIn"] = new_contained_in
    return json_ld



"""Migrate parameters that don't exist in schema.org."""

import json


def up(json_ld):
    """Performs the following sc->cr migration.

    sc:md5->cr:md5, sc:FileSet->cr:FileSet, sc:FileObject->cr:FileObject, sc:key->cr:key
    """
    json_ld = json.dumps(json_ld)
    json_ld = json_ld.replace("sc:FileObject", "cr:FileObject")
    json_ld = json_ld.replace("sc:FileSet", "cr:FileSet")
    json_ld = json.loads(json_ld)
    json_ld["@context"]["key"] = "cr:key"
    json_ld["@context"]["md5"] = "cr:md5"
    return json_ld



"""Tests for the loading script."""

from etils import epath
import pytest

from mlcroissant._src.tests.versions import parametrize_version
from mlcroissant.scripts import load as load_lib


def test_should_raise_when_no_record_set():
    dataset_name = "huggingface-mnist"
    file = (
        epath.Path(__file__).parent.parent.parent.parent.parent
        / "datasets"
        / "1.0"
        / dataset_name
        / "metadata.json"
    )
    with pytest.raises(
        ValueError,
        match="--record_set flag should have a value in `mnist_splits`, `default`",
    ):
        load_lib.load(jsonld=file, record_set=None)


@parametrize_version()
def test_should_raise_when_invalid_mapping(version):
    file = (
        epath.Path(__file__).parent.parent.parent.parent.parent
        / "datasets"
        / version
        / "huggingface-mnist"
        / "metadata.json"
    )
    with pytest.raises(
        ValueError, match="--mapping should be a valid dict\\[str, str\\]"
    ):
        load_lib.load(jsonld=file, record_set="default", mapping="foobarbaz")



"""Tests for RecordSets."""

from unittest import mock

import pytest

from mlcroissant._src.core import constants
from mlcroissant._src.core.context import Context
from mlcroissant._src.core.context import CroissantVersion
from mlcroissant._src.structure_graph.base_node import Node
from mlcroissant._src.structure_graph.nodes.field import Field
from mlcroissant._src.structure_graph.nodes.record_set import get_parent_uuid
from mlcroissant._src.structure_graph.nodes.record_set import RecordSet
from mlcroissant._src.tests.nodes import create_test_field
from mlcroissant._src.tests.nodes import create_test_node
from mlcroissant._src.tests.nodes import create_test_record_set


@pytest.mark.parametrize(
    ["data", "error"],
    [
        [
            {"foo": "bar"},
            (
                "[RecordSet(record_set_name)] http://mlcommons.org/croissant/data"
                " should declare a list. Got: <class 'dict'>."
            ),
        ],
        [
            [],
            (
                "[RecordSet(record_set_name)] http://mlcommons.org/croissant/data"
                " should declare a non empty list."
            ),
        ],
        [
            [[{"foo": "bar"}]],
            (
                "[RecordSet(record_set_name)] http://mlcommons.org/croissant/data"
                " should declare a list of dict. Got: a list of <class 'list'>."
            ),
        ],
        [
            [{"foo": "bar"}],
            (
                "[RecordSet(record_set_name)] Line #0 doesn't have the expected"
                " columns. Expected: {'field_name'}. Got: {'foo'}."
            ),
        ],
    ],
)
def test_invalid_data(data, error):
    ctx = Context()
    field = create_test_field(ctx=ctx)
    create_test_record_set(
        ctx=ctx,
        data=data,
        fields=[field],
    )
    assert error in ctx.issues.errors


@pytest.mark.parametrize(
    ["conforms_to", "field_uuid"],
    [[CroissantVersion.V_0_8, "name"], [CroissantVersion.V_1_0, "id"]],
)
def test_checks_are_performed(conforms_to, field_uuid):
    with mock.patch.object(
        Node, "assert_has_mandatory_properties"
    ) as mandatory_mock, mock.patch.object(Node, "validate_name") as validate_name_mock:
        ctx = Context(conforms_to=conforms_to)
        create_test_node(RecordSet, ctx=ctx)
        mandatory_mock.assert_called_once_with(field_uuid)
        validate_name_mock.assert_called_once()


def test_from_jsonld():
    ctx = Context()
    jsonld = {
        "@type": constants.ML_COMMONS_RECORD_SET_TYPE(ctx),
        "@id": "foo_id",
        constants.SCHEMA_ORG_NAME: "foo",
        constants.SCHEMA_ORG_DESCRIPTION: "bar",
        constants.ML_COMMONS_IS_ENUMERATION(ctx): True,
        constants.SCHEMA_ORG_KEY(ctx): [{"@id": "key1"}, {"@id": "key2"}],
        constants.ML_COMMONS_DATA(ctx): '[{"column1": ["value1", "value2"]}]',
    }
    record_set = RecordSet.from_jsonld(ctx, jsonld)
    assert record_set.name == "foo"
    assert record_set.description == "bar"
    assert record_set.is_enumeration == True
    assert record_set.key == ["key1", "key2"]
    assert record_set.data == [{"column1": ["value1", "value2"]}]
    assert ctx.issues.errors == {
        (
            "[RecordSet(foo)] Line #0 doesn't have the expected columns. Expected:"
            " set(). Got: {'column1'}."
        ),
        (
            "[RecordSet(foo)] Line #0 doesn't have the expected columns. Expected:"
            " set(). Got: {'column1'}."
        ),
    }


@pytest.mark.parametrize(
    ["node_type", "uuid", "parent_uuid", "input", "output"],
    [
        [RecordSet, "foo", "other", "foo", "foo"],
        [Field, "foo/bar", "foo", "foo/bar", "foo"],
    ],
)
def test_get_parent_uuid(node_type, uuid, parent_uuid, input, output):
    mocked_node = mock.Mock(uuid=uuid, spec_set=node_type)
    mocked_node.parent.uuid = parent_uuid
    mocked_ctx = Context()
    mocked_ctx.graph.nodes = lambda: [mocked_node]
    assert get_parent_uuid(mocked_ctx, input) == output



"""Migrates Croissant configs.

Migration from an older or non canonical Croissant format to a canonical and possibly
newer Croissant format.

## What the migration script does

- Read the current file.
- Possibly apply a custom `up` function.
- Add the `@context` defined in `mlcroissant/_src/core/rdf.py`.
- Re-compacting back the Croissant file.

## What you have to do

- If you want to change the `@context` in Croissant files. Then change the context in
the `make_context` function in mlcroissant/_src/core/rdf.py and launch the migration:

```bash
python mlcroissant/scripts/migrations/migrate.py
```

- If you want to migrate a property in every file, you'll need to write a custom
migration function. You can do this by writing a file like
in `previous/YYYYmmddHHmm.py` (similar to previous/202307171508.py) that defines a `up`
function.

```bash
python mlcroissant/scripts/migrations/migrate.py --migration 202307171508
```

Committing your migration allows to keep track of previous migrations in the codebase.
"""

import importlib
import json
import os

from absl import app
from absl import flags
from etils import epath

import mlcroissant as mlc
from mlcroissant._src.core.context import Context
from mlcroissant._src.core.rdf import make_context

_PREVIOUS_MIGRATIONS_FOLDER = "previous"

flags.DEFINE_string(
    "migration",
    None,
    "The name of the Python file with the migration.",
)

flags.DEFINE_string(
    "version",
    "1.0",
    "The version to migrate (datasets/0.8, datasets/1.0, etc).",
)

FLAGS = flags.FLAGS


def get_migration_fn(migration: str | None):
    """Retrieves the `up` migration function from the migration file if it exists."""
    if migration is None:

        def identity_function(x):
            return x

        return identity_function
    try:
        mod = importlib.import_module(f"{_PREVIOUS_MIGRATIONS_FOLDER}.{migration}")
    except ImportError as e:
        raise ValueError(
            f"Did you create a file named {_PREVIOUS_MIGRATIONS_FOLDER}/{migration}.py?"
        ) from e
    try:
        return getattr(mod, "up")
    except AttributeError as e:
        raise ValueError(
            f"Does the file {_PREVIOUS_MIGRATIONS_FOLDER}/{migration}.py declare a `up`"
            " function?"
        ) from e


def migrate_dataset(json_ld):
    """Migrates a regular Croissant file using mlcroissant Python API."""
    metadata = mlc.Metadata.from_json(ctx=Context(), json_=json_ld)
    return metadata.to_json()


def migrate_test_dataset(dataset: epath.Path, json_ld):
    """Migrates a test Croissant files.

    Cannot use mlc.Metadata as test Croissant files may contain errors.
    """
    # Here you can process json_ld as such:
    # ```
    # from mlcroissant._src.core.json_ld import compact_jsonld
    # from mlcroissant._src.core.json_ld import expand_jsonld
    # try:
    #     json_ld = compact_jsonld(expand_jsonld(json_ld, ctx=Context()))
    # except mlc.ValidationError:
    #     pass
    # ```
    # But this behaviour is still bugged, so for now you have to perform all
    # transformations in the up migration function.
    if "@base" in json_ld["@context"]:
        del json_ld["@context"]["@base"]
    # Special cases for test datasets without @context
    if "recordset_missing_context_for_datatype" in os.fspath(dataset):
        del json_ld["@context"]["dataType"]
    if "mlfield_missing_source" in os.fspath(dataset):
        del json_ld["@context"]["source"]
    return json_ld


def standardize_context(json_ld):
    """Standardizes the @context by merging it with the actual context."""
    existing_context = json_ld["@context"]
    context = make_context() | existing_context
    json_ld["@context"] = context
    return json_ld


def main(argv):
    """Main function launched for the migration."""
    del argv
    version = FLAGS.version
    # Datasets in croissant/datasets
    datasets_path = (
        epath.Path(__file__).parent.parent.parent.parent.parent.parent
        / "datasets"
        / version
    )
    datasets = [path for path in datasets_path.glob("*/*.json")]
    assert datasets, f"No dataset found in {datasets_path}"
    # Datasets in croissant/python/mlcroissant/_src/tests
    test_path = (
        epath.Path(__file__).parent.parent.parent.parent
        / "mlcroissant/_src/tests/graphs"
        / version
    )
    test_datasets = [
        p
        for p in test_path.glob("*/*.json")
        if not os.fspath(p).endswith("recordset_bad_type/metadata.json")
    ]
    assert test_datasets, f"No dataset found in {test_path}"
    up = get_migration_fn(FLAGS.migration)
    for dataset in datasets:
        print(f"Converting {dataset}...")
        with dataset.open("r") as f:
            json_ld = json.load(f)
            json_ld = up(json_ld)
        json_ld = migrate_dataset(json_ld)
        json_ld = standardize_context(json_ld)
        with dataset.open("w") as f:
            json.dump(json_ld, f, indent="  ")
            f.write("\n")
    for dataset in test_datasets:
        print(f"Converting test dataset {dataset}...")
        if "mlfield_bad_type" in str(dataset):
            with dataset.open("r") as f:
                json_ld = json.load(f)
                json_ld = up(json_ld)
            json_ld = standardize_context(json_ld)
            json_ld = migrate_test_dataset(dataset, json_ld)
            with dataset.open("w") as f:
                json.dump(json_ld, f, indent="  ")
                f.write("\n")
    print("Done.")


if __name__ == "__main__":
    app.run(main)



"""Test PyTorch utilities in torch modules."""

import pytest

import mlcroissant._src.torch.torch_adapter.dataloader as dataloader


@pytest.mark.parametrize(
    ["in_val", "expected_val", "data_type"],
    [
        ["Nos vemos mañana".encode("utf-8"), "Nos vemos mañana".encode("utf-8"), None],
        [
            "Nos vemos mañana".encode("utf-8"),
            "Nos vemos mañana",
            dataloader.LoaderSpecificationDataType.INFER,
        ],
        [
            "Nos vemos mañana".encode("utf-8"),
            "Nos vemos mañana",
            dataloader.LoaderSpecificationDataType.UTF8,
        ],
        [1, 1, dataloader.LoaderSpecificationDataType.INFER],
        [1, 1, None],
        [None, None, dataloader.LoaderSpecificationDataType.INFER],
        [None, None, None],
    ],
)
def test_data_transformation_helper(in_val, expected_val, data_type):
    new_val = dataloader.apply_data_type_transformation(in_val, data_type)
    assert new_val == expected_val



"""datasets module."""

from __future__ import annotations

from collections.abc import Mapping
import dataclasses
from typing import Any

from absl import logging
from etils import epath
import networkx as nx

from mlcroissant._src.core import regex as regex_lib
from mlcroissant._src.core.context import Context
from mlcroissant._src.core.graphs import utils as graphs_utils
from mlcroissant._src.core.issues import ValidationError
from mlcroissant._src.operation_graph import OperationGraph
from mlcroissant._src.operation_graph.base_operation import Operation
from mlcroissant._src.operation_graph.base_operation import Operations
from mlcroissant._src.operation_graph.execute import execute_downloads
from mlcroissant._src.operation_graph.execute import execute_operations_in_beam
from mlcroissant._src.operation_graph.execute import execute_operations_in_streaming
from mlcroissant._src.operation_graph.execute import execute_operations_sequentially
from mlcroissant._src.operation_graph.operations import FilterFiles
from mlcroissant._src.operation_graph.operations import InitOperation
from mlcroissant._src.operation_graph.operations import ReadFields
from mlcroissant._src.structure_graph.nodes.field import Field
from mlcroissant._src.structure_graph.nodes.metadata import Metadata
from mlcroissant._src.structure_graph.nodes.source import FileProperty

Filters = Mapping[str, Any]


def get_operations(ctx: Context, metadata: Metadata) -> OperationGraph:
    """Returns operations from the metadata."""
    operations = OperationGraph.from_nodes(ctx=ctx, metadata=metadata)
    operations.check_graph(ctx=ctx)
    if ctx.issues.errors:
        raise ValidationError(ctx.issues.report())
    elif ctx.issues.warnings:
        logging.warning(ctx.issues.report())
    return operations


def _check_mapping(metadata: Metadata, mapping: Mapping[str, epath.Path]):
    """Checks that the mapping is valid, i.e. keys are actual UUIDs and paths exist."""
    uuids = set([node.uuid for node in metadata.nodes()])
    for uuid, path in mapping.items():
        if uuid not in uuids:
            raise ValueError(f"{uuid=} in the mapping doesn't exist in the JSON-LD")
        if not path.exists():
            raise ValueError(f"{path=} doesn't exist on disk")


def _expand_mapping(
    mapping: Mapping[str, epath.PathLike] | None,
) -> Mapping[str, epath.Path]:
    """Expands the file mapping to pathlib-readable paths."""
    if isinstance(mapping, Mapping):
        return {key: epath.Path(value).expanduser() for key, value in mapping.items()}
    return {}


@dataclasses.dataclass
class Dataset:
    """Python representation of a Croissant dataset.

    Args:
        jsonld: A JSON object or a path to a Croissant file (URL, str or pathlib.Path).
        debug: Whether to print debug hints. False by default.
        mapping: Mapping filename->filepath as a Python dict[str, str] to handle manual
            downloads. If `document.csv` is the FileObject and you downloaded it to
            `~/Downloads/document.csv`, you can specify `mapping={"document.csv":
            "~/Downloads/document.csv"}`.
    """

    jsonld: epath.PathLike | str | dict[str, Any] | None
    operations: OperationGraph = dataclasses.field(init=False)
    metadata: Metadata = dataclasses.field(init=False)
    debug: bool = False
    mapping: Mapping[str, epath.PathLike] | None = None

    def __post_init__(self):
        """Runs the static analysis of `file`."""
        ctx = Context(mapping=_expand_mapping(self.mapping))
        if isinstance(self.jsonld, dict):
            self.metadata = Metadata.from_json(ctx=ctx, json_=self.jsonld)
        elif self.jsonld is not None:
            self.metadata = Metadata.from_file(ctx=ctx, file=self.jsonld)
        else:
            return
        _check_mapping(self.metadata, ctx.mapping)
        # Draw the structure graph for debugging purposes.
        if self.debug:
            graphs_utils.pretty_print_graph(ctx.graph)
        self.operations = get_operations(ctx, self.metadata)
        # Draw the operations graph for debugging purposes.
        if self.debug:
            graphs_utils.pretty_print_graph(self.operations.operations)

    @classmethod
    def from_metadata(cls, metadata: Metadata) -> Dataset:
        """Creates a new `Dataset` from a `Metadata`."""
        dataset = Dataset(jsonld=None)
        dataset.metadata = metadata
        dataset.operations = get_operations(metadata.ctx, metadata)
        return dataset

    def records(self, record_set: str, filters: Filters | None = None) -> Records:
        """Accesses all records with @id==record_set if it exists.

        record_set: The name of the record set to access.
        filters: A dictionary mapping a field ID to the value we want to filter in. For
            example, when writing {'data/split': 'train'}, we want to keep all records
            whose field `data/split` takes the value `train`.
        """
        if filters:
            _validate_filters(filters)

        if not any(rs for rs in self.metadata.record_sets if rs.uuid == record_set):
            ids = [record_set.uuid for record_set in self.metadata.record_sets]
            error_msg = f"did not find any record set with the name `{record_set}`. "
            if not ids:
                error_msg += "This dataset declares no record sets."
            else:
                error_msg += f"Possible RecordSets: {ids}"
            raise ValueError(error_msg)
        return Records(
            dataset=self,
            record_set=record_set,
            filters=filters,
            debug=self.debug,
        )


@dataclasses.dataclass(kw_only=True)
class Records:
    """Iterable set of records.

    Args:
        dataset: The parent dataset.
        record_set: The name of the record set.
        debug: Whether to print debug hints.
    """

    dataset: Dataset
    record_set: str
    filters: Filters | None
    debug: bool

    def __iter__(self):
        """Executes all operations, runs dynamic analysis and yields examples.

        Warning: at the moment, this method yields examples from the first explored
        record_set.
        """
        # We only consider the operations that are useful to produce the `ReadFields`.
        operations = self._filter_interesting_operations(self.filters)
        # Downloads can be parallelized, so we execute them in priority.
        execute_downloads(operations)
        # We can stream the dataset iff the operation graph is a path graph (meaning
        # that all operations lie on a single straight line, i.e. have an
        # in-degree of 0 or 1. That means that the operation graph is a single line
        # (without external joins for example).
        if _is_streamable_dataset(operations):
            yield from execute_operations_in_streaming(
                record_set=self.record_set,
                operations=operations,
            )
        else:
            yield from execute_operations_sequentially(
                record_set=self.record_set, operations=operations
            )

    def beam_reader(self):
        """See ReadFromCroissant docstring."""
        operations = self._filter_interesting_operations(self.filters)
        execute_downloads(operations)
        return execute_operations_in_beam(
            record_set=self.record_set,
            operations=operations,
            filters=self.filters,
        )

    def _filter_interesting_operations(self, filters: Filters | None) -> Operations:
        """Filters connected operations to `ReadFields(self.record_set)`.

        This function does 2 things:

        - It keeps only the operations that are necessary to produce the record set.
        - It applies the filters to all operations. Example: if we request the training
          set, it will only download the data relative to the training set.
        """
        operations = self.dataset.operations.operations
        source = next(
            operation
            for operation in operations.nodes
            if isinstance(operation, InitOperation)
        )
        target = next(
            operation
            for operation in operations.nodes
            if isinstance(operation, ReadFields)
            and operation.node.uuid == self.record_set
        )
        paths = nx.all_simple_paths(operations, source=source, target=target)
        interesting_nodes = {node for path in paths for node in path}
        interesting_operations = operations.subgraph(interesting_nodes)
        if not filters:
            return interesting_operations  # pytype: disable=bad-return-type
        field, value = _find_data_field_to_filter(filters, interesting_operations)
        new_regex = _regex_from_value(field, value)
        _propagate_includes(field, interesting_operations, new_regex)
        # The value of `field` is now entirely known so we can remove any operation
        # needed to compute it, i.e. all operations involved in a potential join:
        join_uuid = field.references.uuid
        graph = field.ctx.graph
        if join_uuid:
            join_node = next(
                node
                for node in graph
                if isinstance(node, Field) and node.uuid == join_uuid
            )
            unneeded_nodes = [
                node
                for node in graph
                if graph.has_edge(node, join_node) or node == join_node
            ]
            interesting_operations = [
                o for o in interesting_operations if o.node not in unneeded_nodes
            ]
        return operations.subgraph(
            interesting_operations
        )  # pytype: disable=bad-return-type


def _find_data_field_to_filter(
    filters: Filters, operations: nx.Graph[Operation]
) -> tuple[Field, Any]:
    for operation in operations:
        if isinstance(operation, ReadFields):
            for field in operation.node.fields:
                if field.uuid in filters:
                    return field, filters[field.id]
    raise ValueError(
        f"Filters ({filters}) do not apply to the fields. `filters` must be a"
        " mapping from the field ID to the value to filter in, e.g. `{'split':"
        " 'train'}` to filter the train split."
    )


def _regex_from_value(field: Field, value: Any):
    """Creates a regular expression by injecting the value in the transformation."""
    transforms = field.source.transforms
    error = (
        "Filtering is currently only implemented on fields with one transformation from"
        " a regex."
    )
    if len(transforms) != 1:
        raise NotImplementedError(error)
    transform = transforms[0]
    if str_regex := transform.regex:
        return regex_lib.capture_one_capturing_group(str_regex, value)
    raise NotImplementedError(error)


def _propagate_includes(field: Field, operations: nx.Graph[Operation], new_regex: str):
    """Propagates the new_regex back to the source includes.

    Warning: the following lines are a heuristic to inject the regex in the glob
    pattern. First we split the glob pattern by `/` hoping that the filename glob
    pattern comes after this `/`. Here is an example, with the name of the variables and
    a possible example:

    pattern   ->  filename_pattern  ->  filename   ->  new_pattern
    **/*.jpg  ->  *.jpg             ->  train.jpg  ->  **/train.jpg

    Then, we update in place the nodes with the new pattern.
    """
    source_uuid = field.source.uuid
    error = (
        "Currently, filtering is only implemented when extracting a"
        " field from its filename or fullpath. Please, open a GitHub"
        " issue if you want support for other extract."
    )
    source_type = field.source.extract.file_property
    if not field.source.extract.file_property:
        raise NotImplementedError(error)
    for operation in operations:
        if isinstance(operation, FilterFiles):
            node = operation.node
            if node.uuid == source_uuid and new_regex:
                includes = node.includes or []
                if source_type == FileProperty.filename:
                    new_includes = []
                    for pattern in includes:
                        filename_pattern = pattern.split("/")
                        if len(filename_pattern) <= 1:
                            raise NotImplementedError()
                        filenames = regex_lib.regex_to_glob(new_regex)
                        for filename in filenames:
                            new_pattern = filename_pattern[:-1] + [filename]
                            new_includes.append("/".join(new_pattern))
                    node.includes = new_includes
                elif source_type == FileProperty.fullpath:
                    node.includes = regex_lib.regex_to_glob(new_regex)
                else:
                    raise NotImplementedError(error)


def _validate_filters(filters: Filters):
    if isinstance(filters, Mapping) and len(filters) <= 1:
        if all(isinstance(value, str) for value in filters.values()):
            return
    raise ValueError(
        "Filters should be a mapping from a field's ID to the value we want to"
        " filter in. For example, when writing {'data/split': 'train'}, we want"
        " to keep all records whose field `data/split` takes the value `train`."
        f" Instead, we got: {filters=}"
    )


def _is_streamable_dataset(operations: Operations):
    """Whether the operations define a streamable datasets.

    A streamable dataset is a dataset that results from executing a linear sequence of
    operations without branching (for example, no join).
    """
    return all(d == 1 or d == 2 for _, d in operations.degree())



"""Migration: Uniforme names and ids.

In fields and record sets, make the names the same as the ids, as it is mostly done in
the specs.
"""


def up(json_ld):
    """Up function."""
    json_copy = json_ld.copy()
    for _, record_set in enumerate(json_copy.get("recordSet", [])):
        if record_set["@id"] != record_set["name"]:
            record_set["name"] = record_set["@id"]
        fields = record_set.get("field", [])
        for _, field in enumerate(fields):
            if field["@id"] != field["name"]:
                field["name"] = field["@id"]
    return json_ld



"""datasets_test module."""

import json
from typing import Any

from apache_beam.testing import test_pipeline
from apache_beam.testing.util import assert_that
from etils import epath
import pytest

from mlcroissant._src import datasets
from mlcroissant._src.beam import ReadFromCroissant
from mlcroissant._src.core.issues import ValidationError
from mlcroissant._src.core.optional import deps
from mlcroissant._src.tests.records import record_to_python
from mlcroissant._src.tests.versions import parametrize_version

_REPOSITORY_FOLDER = epath.Path(__file__).parent.parent.parent.parent.parent


# End-to-end tests on real data. The data is in `tests/graphs/*/metadata.json`.
def get_error_msg(folder: epath.Path):
    path = folder / "output.txt"
    with path.open("r") as file:
        return file.read().strip()


@parametrize_version()
@pytest.mark.parametrize(
    "folder",
    [
        # Distribution.
        "distribution_bad_contained_in",
        "distribution_bad_type",
        "distribution_missing_encoding_format",
        "distribution_missing_property_content_url",
        # Metadata.
        "metadata_bad_type",
        # ML field.
        "mlfield_bad_source",
        "mlfield_bad_type",
        "mlfield_missing_source",
        # Record set.
        "recordset_bad_type",
        "recordset_missing_context_for_datatype",
        "recordset_wrong_join",
    ],
)
def test_static_analysis(version, folder):
    base_path = epath.Path(__file__).parent / "tests/graphs" / version
    with pytest.raises(ValidationError) as error_info:
        datasets.Dataset(base_path / f"{folder}/metadata.json")
    assert str(error_info.value) == get_error_msg(base_path / folder)


# These tests refer to properties which were mandatory for Croissant 0.8, but not 1.0.
@pytest.mark.parametrize(
    "folder",
    [
        "distribution_missing_name",
        "metadata_missing_property_name",
        "mlfield_missing_property_name",
        "recordset_missing_property_name",
    ],
)
def test_static_analysis_0_8(folder):
    base_path = epath.Path(__file__).parent / "tests/graphs" / "0.8"
    with pytest.raises(ValidationError) as error_info:
        datasets.Dataset(base_path / f"{folder}/metadata.json")
    assert str(error_info.value) == get_error_msg(base_path / folder)


# Tests for 1.0-datasets only.
@pytest.mark.parametrize(
    "folder",
    [
        "distribution_bad_id",
    ],
)
def test_static_analysis_1_0(folder):
    base_path = epath.Path(__file__).parent / "tests/graphs/1.0"
    with pytest.raises(ValidationError) as error_info:
        datasets.Dataset(base_path / f"{folder}/metadata.json")
    assert str(error_info.value) == get_error_msg(base_path / folder)


# Tests for 1.1-datasets only.
@pytest.mark.parametrize(
    "folder",
    [
        "mlfield_bad_array_definition",
        "mlfield_bad_array_shape",
    ],
)
def test_static_analysis_1_1(folder):
    base_path = epath.Path(__file__).parent / "tests/graphs/1.1"
    with pytest.raises(ValidationError) as error_info:
        datasets.Dataset(base_path / f"{folder}/metadata.json")
    assert str(error_info.value) == get_error_msg(base_path / folder)


def load_records_and_test_equality(
    version: str,
    dataset_name: str,
    record_set_name: str,
    num_records: int,
    filters: dict[str, Any] | None = None,
    mapping: dict[str, epath.PathLike] | None = None,
):
    filters_command = ""
    if filters:
        filters_command = str(filters).replace("'", '"')
        filters_command = f" --filters '{filters_command}'"
    print(
        "If this test fails, update JSONL with: `mlcroissant load --jsonld"
        f" ../../datasets/{version}/{dataset_name} --record_set"
        f" {record_set_name} --num_records {num_records} --debug --update_output"
        f" {filters_command}`"
    )
    if dataset_name.startswith("https://"):
        config = dataset_name
        expected_records = None
    else:
        config = _REPOSITORY_FOLDER / "datasets" / version / dataset_name
        output_file = config.parent / "output" / f"{record_set_name}.jsonl"
        with output_file.open("rb") as f:
            lines = f.readlines()
            expected_records = [json.loads(line) for line in lines]
            if num_records > 0:
                assert len(expected_records) == num_records
    dataset = datasets.Dataset(config, mapping=mapping)
    records = dataset.records(record_set_name, filters=filters)
    records = iter(records)
    for i in range(num_records):
        record = next(records)
        if num_records > 0 and i >= num_records:
            break
        record = record_to_python(record)
        if expected_records:
            assert record == expected_records[i]


def _equal_to_set(expected):
    """Checks whether 2 beam.PCollections are equal as sets."""

    def matcher_fn(actual):
        # Sort by index, then remove the index from the PCollection returned by Beam:
        actual = [element for _, element in sorted(list(actual))]
        expected_set = set([
            json.dumps(record_to_python(element)) for element in list(expected)
        ])
        actual_set = set([
            json.dumps(record_to_python(element)) for element in list(actual)
        ])
        assert expected_set == actual_set

    return matcher_fn


def load_records_with_beam_and_test_equality(
    version: str,
    dataset_name: str,
    record_set_name: str,
):
    jsonld = _REPOSITORY_FOLDER / "datasets" / version / dataset_name
    output_file = jsonld.parent / "output" / f"{record_set_name}.jsonl"
    with output_file.open("rb") as f:
        lines = f.readlines()
        expected_records = [json.loads(line) for line in lines]

    with test_pipeline.TestPipeline() as pipeline:
        result = pipeline | ReadFromCroissant(jsonld=jsonld, record_set=record_set_name)
        assert_that(result, _equal_to_set(expected_records))


# IF (NON)-HERMETIC TESTS FAIL, OR A NEW DATASET IS ADDED:
# You can regenerate .pkl files by launching
# ```bash
# mlcroissant load \
#   --jsonld ../../datasets/{{version}}/{{dataset_name}}/metadata.json \
#   --record_set {{record_set_name}} \
#   --update_output \
#   --num_records -1
# ```


# Hermetic test cases (data from local folders).
@parametrize_version()
@pytest.mark.parametrize(
    ["dataset_name", "record_set_name", "num_records"],
    [
        ["audio_test/metadata.json", "records", 2],
        ["pass-mini/metadata.json", "images", -1],
        ["recipes/file_object_in_zip.json", "csv1", -1],
        ["recipes/file_object_in_zip.json", "csv2", -1],
        ["recipes/read_binary_file_by_line.json", "translations_from_directory", -1],
        ["recipes/read_binary_file_by_line.json", "translations_from_zip", -1],
        ["recipes/read_from_directory.json", "read_from_directory_example", -1],
        ["recipes/read_from_tar.json", "images_with_annotations", -1],
        ["simple-join/metadata.json", "publications_by_user", -1],
        ["simple-parquet/metadata.json", "persons", -1],
    ],
)
def test_hermetic_loading(version, dataset_name, record_set_name, num_records):
    load_records_and_test_equality(version, dataset_name, record_set_name, num_records)


@parametrize_version()
@pytest.mark.parametrize(
    ["dataset_name", "record_set_name"],
    [
        ["simple-parquet/metadata.json", "persons"],
        ["simple-join/metadata.json", "publications_by_user"],
    ],
)
def test_beam_hermetic_loading(version, dataset_name, record_set_name):
    load_records_with_beam_and_test_equality(version, dataset_name, record_set_name)


# Hermetic test cases for croissant >=1.0 only.
@pytest.mark.parametrize(
    ["dataset_name", "record_set_name", "num_records", "filters"],
    [
        ["coco2014-mini/metadata.json", "bounding_boxes", -1, None],
        ["coco2014-mini/metadata.json", "captions", -1, None],
        ["coco2014-mini/metadata.json", "images", -1, None],
        ["coco2014-mini/metadata.json", "split_enums", -1, None],
        ["simple-split/metadata.json", "data", -1, {"data/split": "train"}],
    ],
)
def test_hermetic_loading_1_0(dataset_name, record_set_name, num_records, filters):
    load_records_and_test_equality(
        "1.0", dataset_name, record_set_name, num_records, filters
    )


# Non-hermetic test cases (data from the internet).
@pytest.mark.nonhermetic
@parametrize_version()
@pytest.mark.parametrize(
    ["dataset_name", "record_set_name", "num_records"],
    [
        [
            "flores-200/metadata.json",
            "language_translations_train_data_with_metadata",
            10,
        ],
        [
            "flores-200/metadata.json",
            "language_translations_test_data_with_metadata",
            10,
        ],
        ["gpt-3/metadata.json", "default", 10],
        ["huggingface-mnist/metadata.json", "default", 10],
        ["titanic/metadata.json", "passengers", -1],
    ],
)
def test_nonhermetic_loading(version, dataset_name, record_set_name, num_records):
    load_records_and_test_equality(version, dataset_name, record_set_name, num_records)


# Non-hermetic test cases for croissant >=1.0 only (data from the internet).
@pytest.mark.nonhermetic
@pytest.mark.parametrize(
    ["dataset_name", "record_set_name", "num_records", "filters"],
    [
        ["huggingface-anthropic-hh-rlhf/metadata.json", "red-team-attempts", 10, None],
        ["huggingface-c4/metadata.json", "data", 1, {"data/variant": "en"}],
        ["huggingface-levanti/metadata.json", "levanti_train", 10, None],
        ["huggingface-open-hermes/metadata.json", "default", 3, None],
        # This dataset will timeout if the following feature is broken: mlcroissant
        # yields examples by downloading parquet files one by one. mlcroissant should
        # not download all parquet files upfront.
        [
            "https://huggingface.co/api/datasets/bigcode/the-stack-metadata/croissant",
            "default",
            1,
            {"default/split": "train"},
        ],
    ],
)
def test_nonhermetic_loading_1_0(dataset_name, record_set_name, num_records, filters):
    load_records_and_test_equality(
        "1.0", dataset_name, record_set_name, num_records, filters
    )


@pytest.mark.nonhermetic
def test_load_from_huggingface():
    url = "https://huggingface.co/api/datasets/mnist/croissant"
    dataset = datasets.Dataset(url)
    has_one_record = False
    for record in dataset.records(record_set="mnist"):
        assert record["mnist/label"] == 7
        assert isinstance(record["mnist/image"], deps.PIL_Image.Image)
        has_one_record = True
        break
    assert has_one_record, (
        "mlc.Dataset.records() didn't yield any record. Warning: this test is"
        " non-hermetic and makes an API call to Hugging Face, so it's prone to network"
        " failure."
    )


@parametrize_version()
def test_raises_when_the_record_set_does_not_exist(version):
    dataset_folder = _REPOSITORY_FOLDER / "datasets" / version / "titanic"
    dataset = datasets.Dataset(dataset_folder / "metadata.json")
    with pytest.raises(ValueError, match="did not find"):
        dataset.records("this_record_set_does_not_exist")


@parametrize_version()
def test_cypress_fixtures(version):
    # Cypress cannot read files outside of its direct scope, so we have to copy them
    # as fixtures. This test tests that the copies are equal to the original.
    fixture_folder: epath.Path = (
        _REPOSITORY_FOLDER / "editor" / "cypress" / "fixtures" / version
    )
    datasets_folder: epath.Path = _REPOSITORY_FOLDER / "datasets" / version
    for fixture in fixture_folder.glob("*.json"):
        dataset = datasets_folder / f"{fixture.stem}" / "metadata.json"
        assert json.load(fixture.open()) == json.load(dataset.open()), (
            f"If this test fails, you probably have to copy the content of {dataset} to"
            f" {fixture}. Launch the command `cp {dataset} {fixture}`"
        )


@pytest.mark.parametrize(
    ["filters", "raises"],
    [
        [{}, False],
        [{"split": "test"}, False],
        [{"split": "test", "other_filter": "foo"}, True],
        [{"split": ["train", "test"]}, True],
        [{"split": 1}, True],
    ],
)
def test_validate_filters(filters, raises):
    if raises:
        with pytest.raises(ValueError):
            datasets._validate_filters(filters)
    else:
        datasets._validate_filters(filters)


@parametrize_version()
def test_check_mapping_when_the_key_does_not_exist(version):
    dataset_name = "simple-parquet/metadata.json"
    record_set_name = "persons"
    with pytest.raises(ValueError, match="doesn't exist in the JSON-LD"):
        load_records_and_test_equality(
            version,
            dataset_name,
            record_set_name,
            -1,
            mapping={"this_UUID_does_not_exist": "/this/path/does/not/exist"},
        )


@parametrize_version()
def test_check_mapping_when_the_path_does_not_exist(version):
    dataset_name = "simple-parquet/metadata.json"
    record_set_name = "persons"
    with pytest.raises(ValueError, match="doesn't exist on disk"):
        load_records_and_test_equality(
            version,
            dataset_name,
            record_set_name,
            -1,
            mapping={"dataframe": "/this/path/does/not/exist"},
        )


@parametrize_version()
def test_check_mapping_when_the_mapping_is_correct(version, tmp_path):
    dataset_name = "simple-parquet/metadata.json"
    record_set_name = "persons"
    old_path = (
        _REPOSITORY_FOLDER
        / "datasets"
        / version
        / "simple-parquet/data/dataframe.parquet"
    )
    assert old_path.exists()
    # Copy the dataframe to a temporary file:
    new_path = tmp_path / "dataframe.parquet"
    old_path.copy(new_path)
    load_records_and_test_equality(
        version,
        dataset_name,
        record_set_name,
        -1,
        mapping={"dataframe": new_path},
    )
